from __future__ import annotations

import logging
import typing as t

import attrs
import numpy as np
import pint
import pinttr
import portion as P
import xarray as xr

from .index import CKDSpectralIndex
from ..attrs import documented, parse_docs
from ..constants import SPECTRAL_RANGE_MAX, SPECTRAL_RANGE_MIN
from ..quad import Quad
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg
from ..util.misc import round_to_multiple

logger = logging.getLogger(__name__)

# G16 = Quad.gauss_legendre(16).eval_nodes(interval=[0.0, 1.0])  # TODO: PR#311 hack
# Note: Hard-coded because Mitsuba cannot be imported at module level due to
#       documentation issues.
G16 = np.array(
    [
        0.00529954,
        0.02771249,
        0.06718439,
        0.12229779,
        0.19106188,
        0.27099161,
        0.35919823,
        0.45249375,
        0.54750625,
        0.64080177,
        0.72900839,
        0.80893812,
        0.87770221,
        0.93281561,
        0.97228751,
        0.99470046,
    ]
)


# ------------------------------------------------------------------------------
#                              CKD bin data classes
# ------------------------------------------------------------------------------


@parse_docs
@attrs.define(eq=False, frozen=True, slots=True)
class Bin:
    """
    A data class representing a spectral bin in CKD modes.

    Notes
    -----
    A bin is more than a spectral interval. It is associated with a
    quadrature rule.
    """

    wmin: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("wavelength"),
            on_setattr=None,  # frozen instance: on_setattr must be disabled
        ),
        doc='Bin lower spectral bound.\n\nUnit-enabled field (default: ucc["wavelength"]).',
        type="quantity",
        init_type="quantity or float",
    )

    wmax: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("wavelength"),
            on_setattr=None,  # frozen instance: on_setattr must be disabled
        ),
        doc='Bin upper spectral bound.\n\nUnit-enabled field (default: ucc["wavelength"]).',
        type="quantity",
        init_type="quantity or float",
    )

    @wmin.validator
    @wmax.validator
    def _wbounds_validator(self, attribute, value):
        if not self.wmin < self.wmax:
            raise ValueError(
                f"while validating {attribute.name}: wmin must be lower than wmax"
            )

    quad: Quad = documented(
        attrs.field(
            factory=lambda: Quad.gauss_legendre(2),
            repr=lambda x: x.str_summary,
            validator=attrs.validators.instance_of(Quad),
        ),
        doc="Quadrature rule attached to the CKD bin.",
        type=":class:`.Quad`",
    )

    bin_set_id: str | None = documented(
        attrs.field(default=None, converter=str),
        doc="Id of the bin set used to create this bin.",
        type="str or None",
        init_type="str",
    )

    @property
    def width(self) -> pint.Quantity:
        """quantity : Bin spectral width."""
        return self.wmax - self.wmin

    @property
    def wcenter(self) -> pint.Quantity:
        """quantity : Bin central wavelength."""
        return 0.5 * (self.wmin + self.wmax)

    @property
    def interval(self) -> P.Interval:
        """portion.Interval : Closed-open interval corresponding to the bin
        wavelength interval."""
        return P.closedopen(self.wmin, self.wmax)

    @property
    def pretty_repr(self) -> str:
        """str : Pretty representation of the bin."""
        units = ureg.Unit("nm")
        wrange = (
            f"[{self.wmin.m_as(units):.1f}, {self.wmax.m_as(units):.1f}] {units:~P}"
        )
        quad = self.quad.pretty_repr()
        return f"{wrange}: {quad}"

    def spectral_indices(self) -> t.Generator[CKDSpectralIndex]:
        for value in self.quad.eval_nodes(interval=[0.0, 1.0]):
            yield CKDSpectralIndex(w=self.wcenter, g=value)


# ------------------------------------------------------------------------------
#                              Bin set data class
# ------------------------------------------------------------------------------


@parse_docs
@attrs.define(eq=False, frozen=True, slots=True)
class BinSet:
    """
    A data class representing a bin set used in CKD mode.

    See Also
    --------
    :class:`.WavelengthSet`
    """

    bins: list[Bin] = documented(
        attrs.field(
            converter=list,
            validator=attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(Bin)
            ),
        ),
        doc="Set of bins.",
        type="set of :class:`.Bin`",
        init_type="iterable of :class:`.Bin`",
    )

    def spectral_indices(self) -> t.Generator[CKDSpectralIndex]:
        for bin in self.bins:
            yield from bin.spectral_indices()

    @classmethod
    def arange(
        cls,
        start: pint.Quantity,
        stop: pint.Quantity,
        step: pint.Quantity = 10.0 * ureg.nm,
        quad: Quad | None = None,
    ) -> BinSet:
        """
        Generate a bin set with linearly spaced bins.

        Parameters
        ----------
        start : quantity
            Lower bound of first bin.

        stop : quantity
            Upper bound of last bin.

        step : quantity
            Bin width.

        quad : .Quad, optional
            Quadrature rule (same for all bins in the set). Defaults to
            a two-point Gauss-Legendre quadrature.

        Returns
        -------
        :class:`.BinSet`
            Generated bin set.
        """
        if quad is None:
            quad = Quad.gauss_legendre(2)

        bins = []

        wunits = ucc.get("wavelength")
        _start = start.m_as(wunits)
        _stop = stop.m_as(wunits)
        _step = step.m_as(wunits)

        for wmin in np.arange(_start, _stop, _step):
            wmax = wmin + _step
            bins.append(
                Bin(
                    wmin=wmin * wunits,
                    wmax=wmax * wunits,
                    quad=quad,
                )
            )

        return cls(bins)

    @classmethod
    def from_wavelengths(
        cls,
        wavelengths: pint.Quantity,
        width: pint.Quantity = 10.0 * ureg.nm,
        quad: Quad | None = None,
    ) -> BinSet:
        """
        Generate a bin set with bins centered on the given wavelengths.

        Parameters
        ----------
        wavelengths : sequence of quantity
            Wavelengths to center bins on.

        quad : :class:`.Quad`, optional
            Quadrature rule (same for all bins in the set). Defaults to
            a two-point Gauss-Legendre quadrature.

        Returns
        -------
        :class:`.BinSet`
            Generated bin set.
        """
        if quad is None:
            quad = Quad.gauss_legendre(2)

        bins = []

        for wcenter in np.atleast_1d(wavelengths):
            bins.append(
                Bin(
                    wmin=wcenter - width / 2,
                    wmax=wcenter + width / 2,
                    quad=quad,
                )
            )

        return cls(bins)

    @classmethod
    def from_absorption_dataset(
        cls,
        dataset: xr.Dataset,
        quad: Quad | None = None,
    ) -> BinSet:
        """
        Generate a bin set from an absorption dataset.

        Parameters
        ----------
        dataset : Dataset
            Absorption dataset.

        quad : :class:`.Quad`, optional
            Quadrature rule (same for all bins in the set). Defaults to
            an eight-point Gauss-Legendre quadrature.

        Returns
        -------
        :class:`.BinSet`
            Generated bin set.

        Notes
        -----
        Assumes that:

        * the absorption dataset has a ``bin`` coordinate with values in
          nanometers.
        * the absorption dataset has a ``bin_set`` attribute with the bin
          width in nanometers.
        """
        if quad is None:
            quad = Quad.gauss_legendre(8)

        # TODO: PR#311 hack (next 2 lines)
        wavelengths = np.unique(np.array(dataset.bin.values, dtype=float)) * ureg.nm
        width = ureg(
            dataset.attrs["bin_set"]
        )  # read bin_set dataset attribute and convert it to a quantity
        return cls.from_wavelengths(wavelengths, width=width, quad=quad)

    @classmethod
    def default(cls):
        """
        Generate a default bin set, which covers Eradiate's default spectral
        range with 10 nm-wide bins.
        """
        wmin = round_to_multiple(SPECTRAL_RANGE_MIN.m_as(ureg.nm), 10.0, "nearest")
        wmax = round_to_multiple(SPECTRAL_RANGE_MAX.m_as(ureg.nm), 10.0, "nearest")
        dw = 10.0

        return cls.from_wavelengths(
            wavelengths=np.arange(wmin, wmax + dw, dw) * ureg.nm,
            width=dw * ureg.nm,
            quad=Quad.gauss_legendre(2),
        )



"""Functions to compute Rayleigh scattering in air."""

import numpy as np
import pint
from scipy.constants import physical_constants

from .. import data
from ..units import unit_registry as ureg
from ..util.misc import Singleton

# Physical constants
#: Loschmidt constant [km^-3].
_LOSCHMIDT = ureg.Quantity(
    *physical_constants["Loschmidt constant (273.15 K, 101.325 kPa)"][:2]
).to("km^-3")

# Air number density at 101325 Pa and 288.15 K
_STANDARD_AIR_NUMBER_DENSITY = _LOSCHMIDT * (273.15 / 288.15)


# Bates (1984) King correction factor data
class _BATES_1984_DATA(metaclass=Singleton):
    # Lazy loader for Bates (1984) King correction data
    def __init__(self):
        self._data = None

    @property
    def data(self):
        if self._data is None:
            self._data = data.load_dataset("spectra/optics/bates_1984.nc")
        return self._data


def compute_sigma_s_air(
    wavelength: pint.Quantity = ureg.Quantity(550.0, "nm"),
    number_density: pint.Quantity = _STANDARD_AIR_NUMBER_DENSITY,
) -> pint.Quantity:
    r"""
    Compute the Rayleigh scattering coefficient of air.

    When default values are used, this provides the Rayleigh scattering
    coefficient for air at 550 nm in standard temperature and pressure
    conditions.

    The scattering coefficient is computed by considering the air as a pure
    gas with associated effective optical properties (refractive index,
    King factor) and according to the expression provided by
    :cite:`Eberhard2010CorrectEquationsCommon` (eq. 60):

    .. math::

       k_{\mathrm s \, \lambda} (n) = \frac{8 \pi^3}{3 \lambda^4} \frac{1}{n}
          \left( \eta_{\lambda}^2(n) - 1 \right)^2 F_{\lambda}

    where
    :math:`\lambda` is the wavelength (subscript indicates spectral dependence),
    :math:`n` is the air number density,
    :math:`\eta` is the air refractive index and
    :math:`F` is the air King factor.

    The King correction factor is computed by linearly interpolating the data
    from :cite:`Bates1984RayleighScatteringAir`.

    Parameters
    ----------
    wavelength : quantity
        Wavelength [nm].

    number_density : quantity
        Number density of the scattering particles [km^-3].

    Returns
    -------
    quantity
        Scattering coefficient.
    """
    BATES_1984_DATA = _BATES_1984_DATA().data
    f_left = BATES_1984_DATA.f.values[0]
    f_right = BATES_1984_DATA.f.values[-1]
    king_factor = BATES_1984_DATA.f.interp(
        w=wavelength.m_as("micrometer"),
        kwargs=dict(fill_value=(f_left, f_right)),
    ).values

    refractive_index = air_refractive_index(
        wavelength=wavelength, number_density=number_density
    )
    if isinstance(wavelength.magnitude, np.ndarray) and isinstance(
        number_density.magnitude, np.ndarray
    ):
        king_factor = king_factor[:, np.newaxis]
        wavelength = wavelength[:, np.newaxis]
        number_density = number_density[np.newaxis, :]

    return (
        8.0
        * np.power(np.pi, 3)
        / (3.0 * np.power(wavelength, 4))
        / number_density
        * np.square(np.square(refractive_index) - 1.0)
        * king_factor
    ).to("km^-1")


def air_refractive_index(
    wavelength: pint.Quantity = ureg.Quantity(550.0, "nm"),
    number_density: pint.Quantity = _STANDARD_AIR_NUMBER_DENSITY,
) -> np.ndarray:
    """
    Computes the air refractive index.

    The wavelength dependence of the refractive index is computed using equation
    2 from :cite:`Peck1972DispersionAir`. This formula is a fit of
    measurements of the air refractive index in the range of wavelength from
    :math:`\\lambda = 240` nm to :math:`1690` nm.
    The number density dependence is computed using a simple proportionality
    rule.

    Parameters
    ----------
    wavelength : quantity
        Wavelength.

    number_density : quantity
        Number density.

        Default: Air number density at 101325 Pa and 288.15 K.

    Returns
    -------
    float or ndarray
        Air refractive index value(s).
    """

    # wavenumber in inverse micrometer
    sigma = 1 / wavelength.m_as("micrometer")
    sigma2 = np.square(sigma)

    # refractivity in parts per 1e8
    x = (5791817.0 / (238.0183 - sigma2)) + 167909.0 / (57.362 - sigma2)

    if isinstance(x, np.ndarray) and isinstance(number_density.magnitude, np.ndarray):
        x = x[:, np.newaxis]
        number_density = number_density[np.newaxis, :]

    # number density scaling
    x_scaled = x * (number_density / _STANDARD_AIR_NUMBER_DENSITY).m_as("dimensionless")

    # refractive index
    index = 1 + x_scaled * 1e-8

    return index



import pytest

from eradiate.pipelines import ApplyCallable, Pipeline


def test_pipeline_step_apply_callable():
    f = lambda x: sum(x)

    step = ApplyCallable(f)
    assert step.transform([1, 1]) == 2


def test_pipeline_basic():
    f = lambda x: sum(x)
    g = lambda x: x + 1

    # Init from list of steps
    pipeline = Pipeline([ApplyCallable(f), ApplyCallable(g)])
    assert pipeline.transform([1, 2]) == 4
    assert list(pipeline.named_steps.keys()) == ["0_apply_callable", "1_apply_callable"]
    assert pipeline._names == ["0_apply_callable", "1_apply_callable"]

    # Init with explicit naming
    pipeline = Pipeline([("sum", ApplyCallable(f)), ("add_1", ApplyCallable(g))])
    assert pipeline.transform([1, 2]) == 4
    assert list(pipeline.named_steps.keys()) == ["sum", "add_1"]
    assert pipeline._names == ["sum", "add_1"]


def test_pipeline_add():
    sum_ = ApplyCallable(lambda x: sum(x))
    zero = ApplyCallable(lambda x: 0)
    add_1 = ApplyCallable(lambda x: x + 1)

    # Insert by index
    pipeline = Pipeline([("sum", sum_), ("add_1", add_1)])
    pipeline.add("zero", zero, position=1)
    assert pipeline.transform([1, 2]) == 1

    # Insert before
    pipeline = Pipeline([("sum", sum_), ("add_1", add_1)])
    pipeline.add("zero", zero, before="add_1")
    assert pipeline.transform([1, 2]) == 1

    # Insert after
    pipeline = Pipeline([("sum", sum_), ("add_1", add_1)])
    pipeline.add("zero", zero, after="add_1")
    assert pipeline.transform([1, 2]) == 0

    # Before/after override position
    pipeline = Pipeline([("sum", sum_), ("add_1", add_1)])
    pipeline.add("zero", zero, before="add_1", position=0)
    assert pipeline.transform([1, 2]) == 1

    pipeline = Pipeline([("sum", sum_), ("add_1", add_1)])
    pipeline.add("zero", zero, after="add_1", position=0)
    assert pipeline.transform([1, 2]) == 0

    # Before + after raises
    pipeline = Pipeline([("sum", sum_), ("add_1", add_1)])
    with pytest.raises(ValueError):
        pipeline.add("zero", zero, before="add_1", after="sum")


def test_pipeline_transform():
    add_1 = ApplyCallable(lambda x: x + 1)
    times_3 = ApplyCallable(lambda x: x * 3)
    mod_7 = ApplyCallable(lambda x: x % 7)

    pipeline = Pipeline([("add_1", add_1), ("times_3", times_3), ("mod_7", mod_7)])

    # No param means we execute the whole pipeline
    assert ((3 + 1) * 3) % 7 == pipeline.transform(3)

    # Start index can be used to remove some steps
    assert (3 * 3) % 7 == pipeline.transform(3, start="times_3")
    assert pipeline.transform(3, start="times_3") == pipeline.transform(3, start=1)

    # Stop index as well
    assert (3 + 1) * 3 == pipeline.transform(3, stop="mod_7")
    assert pipeline.transform(3, stop="mod_7") == pipeline.transform(3, stop=2)

    # Stop after index acts like stop but includes its bound
    assert pipeline.transform(3, stop_after="times_3") == pipeline.transform(
        3, stop="mod_7"
    )
    assert pipeline.transform(3, stop_after=1) == pipeline.transform(3, stop=2)
    assert pipeline.transform(3, stop_after=0) == pipeline.transform(3, stop=1)

    # Step selector executes a single step
    assert 3 % 7 == pipeline.transform(3, step="mod_7")
    assert pipeline.transform(3, step="mod_7") == pipeline.transform(3, step=2)



import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.quad import Quad
from eradiate.scenes.spectra import InterpolatedSpectrum, MultiDeltaSpectrum
from eradiate.spectral import Bin, BinSet


def test_ckd_bin():
    """
    Unit tests for :class:`eradiate.ckd.Bin`.
    """
    quad = Quad.gauss_legendre(16)

    # Construct a bin
    bin = Bin(wmin=500.0, wmax=510.0, quad=quad)

    # Scalar values are correctly converted to config default units
    assert bin.wmin == 500.0 * ureg.nm

    # Wrong bound ordering raises
    with pytest.raises(ValueError):
        Bin(wmin=510.0, wmax=500.0, quad=quad)


def test_select_with_multi_delta_1():
    """
    Unit tests for :meth:`.BinSet.select_with`.
    """
    binset = BinSet.arange(
        start=280.0 * ureg.nm,
        stop=2400.0 * ureg.nm,
        step=10.0 * ureg.nm,
    )

    srf = MultiDeltaSpectrum(wavelengths=555.0 * ureg.nm)
    expected_selected_bins = [
        Bin(wmin=550.0 * ureg.nm, wmax=560.0 * ureg.nm),
    ]
    selected = srf.select_in(binset)
    for bin, expected_bin in zip(selected.bins, expected_selected_bins):
        assert np.isclose(bin.wmin, expected_bin.wmin)


def test_select_with_multi_delta_2():
    """
    Unit tests for :meth:`.BinSet.select_with`.
    """
    binset = BinSet.arange(
        start=280.0 * ureg.nm,
        stop=2400.0 * ureg.nm,
        step=10.0 * ureg.nm,
    )

    srf = MultiDeltaSpectrum(wavelengths=550.0 * ureg.nm)
    expected_selected_bins = [
        Bin(wmin=550.0 * ureg.nm, wmax=560.0 * ureg.nm),
    ]
    selected = srf.select_in(binset)
    for bin, expected_bin in zip(selected.bins, expected_selected_bins):
        assert np.isclose(bin.wmin, expected_bin.wmin)


def test_select_from_connex_interpolated_spectrum_1():
    """
    Unit tests for :meth:`.BinSet.select_from_srf`.
    """
    binset = BinSet.arange(
        start=280.0 * ureg.nm,
        stop=2400.0 * ureg.nm,
        step=10.0 * ureg.nm,
    )

    srf = InterpolatedSpectrum(
        wavelengths=np.linspace(500.0, 600.0, 11) * ureg.nm,
        values=np.ones(11),
    )

    selected = srf.select_in(binset)
    wmin_expected = np.linspace(500.0, 600.0, 11) * ureg.nm
    assert len(selected.bins) == 10
    for bin, wmin in zip(selected.bins, wmin_expected):
        assert np.isclose(bin.wmin, wmin)
        assert np.isclose(bin.wmax, wmin + 10.0 * ureg.nm)


def test_select_from_connex_interpolated_spectrum_2():
    """
    Unit tests for :meth:`.BinSet.select_from_srf`.
    """
    binset = BinSet.arange(
        start=280.0 * ureg.nm,
        stop=2400.0 * ureg.nm,
        step=10.0 * ureg.nm,
    )

    srf = InterpolatedSpectrum(
        wavelengths=np.linspace(500.0, 600.0, 11) * ureg.nm,
        values=np.concatenate(
            [
                np.array([0.0]),
                np.ones(9),
                np.array([0.0]),
            ]
        ),
    )

    selected = srf.select_in(binset)
    wmin_expected = np.linspace(500.0, 600.0, 11) * ureg.nm
    assert len(selected.bins) == 10
    for bin, wmin in zip(selected.bins, wmin_expected):
        assert np.isclose(bin.wmin, wmin)
        assert np.isclose(bin.wmax, wmin + 10.0 * ureg.nm)


def test_select_from_connex_interpolated_spectrum_3():
    """
    Unit tests for :meth:`.BinSet.select_from_srf`.
    """
    binset = BinSet.arange(
        start=280.0 * ureg.nm,
        stop=2400.0 * ureg.nm,
        step=10.0 * ureg.nm,
    )

    srf = InterpolatedSpectrum(
        wavelengths=np.linspace(500.0, 600.0, 11) * ureg.nm,
        values=np.concatenate(
            [
                np.array([0.0, 0.0]),
                np.ones(7),
                np.array([0.0, 0.0]),
            ]
        ),
    )

    selected = srf.select_in(binset)
    wmin_expected = np.linspace(510.0, 590.0, 9) * ureg.nm
    assert len(selected.bins) == 8
    for bin, wmin in zip(selected.bins, wmin_expected):
        assert np.isclose(bin.wmin, wmin)
        assert np.isclose(bin.wmax, wmin + 10.0 * ureg.nm)


def test_select_from_non_connex_interpolated_spectrum_1():
    """
    Unit tests for :meth:`.BinSet.select_from_srf`.
    """
    binset = BinSet.arange(
        start=280.0 * ureg.nm,
        stop=2400.0 * ureg.nm,
        step=10.0 * ureg.nm,
    )

    srf = InterpolatedSpectrum(
        wavelengths=np.linspace(500.0, 600.0, 11) * ureg.nm,
        values=np.array([0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0]),
    )

    selected = srf.select_in(binset)
    expected_selected_bins = [
        Bin(wmin=510.0 * ureg.nm, wmax=520.0 * ureg.nm),
        Bin(wmin=520.0 * ureg.nm, wmax=530.0 * ureg.nm),
        Bin(wmin=530.0 * ureg.nm, wmax=540.0 * ureg.nm),
        Bin(wmin=550.0 * ureg.nm, wmax=560.0 * ureg.nm),
        Bin(wmin=560.0 * ureg.nm, wmax=570.0 * ureg.nm),
        Bin(wmin=570.0 * ureg.nm, wmax=580.0 * ureg.nm),
        Bin(wmin=580.0 * ureg.nm, wmax=590.0 * ureg.nm),
    ]
    for bin, expected in zip(selected.bins, expected_selected_bins):
        assert np.isclose(bin.wmin, expected.wmin)
        assert np.isclose(bin.wmax, expected.wmax)


def test_select_from_non_connex_interpolated_spectrum_2():
    """
    Unit tests for :meth:`.BinSet.select_from_srf`.
    """
    binset = BinSet.arange(
        start=280.0 * ureg.nm,
        stop=2400.0 * ureg.nm,
        step=10.0 * ureg.nm,
    )
    srf = InterpolatedSpectrum(
        wavelengths=np.linspace(500.0, 600.0, 11) * ureg.nm,
        values=np.array([0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0]),
    )

    selected = srf.select_in(binset)
    expected_selected_bins = [
        Bin(wmin=510.0 * ureg.nm, wmax=520.0 * ureg.nm),
        Bin(wmin=520.0 * ureg.nm, wmax=530.0 * ureg.nm),
        Bin(wmin=550.0 * ureg.nm, wmax=560.0 * ureg.nm),
        Bin(wmin=560.0 * ureg.nm, wmax=570.0 * ureg.nm),
        Bin(wmin=570.0 * ureg.nm, wmax=580.0 * ureg.nm),
        Bin(wmin=580.0 * ureg.nm, wmax=590.0 * ureg.nm),
        Bin(wmin=590.0 * ureg.nm, wmax=600.0 * ureg.nm),
    ]
    for bin, expected in zip(selected.bins, expected_selected_bins):
        assert np.isclose(bin.wmin, expected.wmin)
        assert np.isclose(bin.wmax, expected.wmax)


def test_select_from_connex_interpolated_spectrum_misaligned():
    """
    Unit tests for :meth:`.BinSet.select_from_srf`.
    """
    binset = BinSet.arange(
        start=280.0 * ureg.nm,
        stop=2400.0 * ureg.nm,
        step=10.0 * ureg.nm,
    )

    srf = InterpolatedSpectrum(
        wavelengths=np.linspace(505.0, 605.0, 11) * ureg.nm,
        values=np.concatenate(
            [
                np.array([0.0, 0.0]),
                np.ones(7),
                np.array([0.0, 0.0]),
            ]
        ),
    )

    selected = srf.select_in(binset)
    wmin_expected = np.linspace(510.0, 600.0, 10) * ureg.nm
    assert len(selected.bins) == 9
    for bin, wmin in zip(selected.bins, wmin_expected):
        assert np.isclose(bin.wmin, wmin)
        assert np.isclose(bin.wmax, wmin + 10.0 * ureg.nm)



"""
Molecular atmospheres.
"""

from __future__ import annotations

import typing as t

import attrs
import numpy as np
import pint
import pinttr
import xarray as xr

import eradiate

from ._core import AbstractHeterogeneousAtmosphere
from ..core import traverse
from ..phase import PhaseFunction, RayleighPhaseFunction, phase_function_factory
from ... import converters, data
from ...attrs import documented, parse_docs
from ...contexts import KernelContext
from ...quad import Quad
from ...radprops import AFGL1986RadProfile, RadProfile, US76ApproxRadProfile, ZGrid
from ...spectral.ckd import BinSet
from ...spectral.index import SpectralIndex
from ...spectral.mono import WavelengthSet
from ...thermoprops import afgl_1986, us76
from ...thermoprops.util import (
    compute_scaling_factors,
    interpolate,
    rescale_concentration,
)
from ...units import unit_registry as ureg
from ...util.misc import summary_repr

DEFAULT_WAVELENGTH_RANGE = [545.0, 555.0] * ureg.nm


def open_us76_dataset(wavelength_range: pint.Quantity) -> xr.Dataset:
    """
    Open the monochromatic absorption dataset corresponding to the given
    wavelength range, for the U.S. 1976 Standard Atmosphere.
    """
    path = "spectra/absorption/us76_u86_4"
    wlmin, wlmax = wavelength_range
    wnmin = (1 / wlmax).m_as("1/cm")
    wnmax = (1 / wlmin).m_as("1/cm")
    w = np.concatenate([np.arange(4000, 25711, 1000), [25711]])
    if wnmin < w[0] or wnmax > w[-1]:
        raise ValueError(
            f"Invalid wavelength range: {wavelength_range}. "
            f"Supported wavelength range: {1e7 / w[-1]:.1f}-"
            f"{1e7 / w[0]:.1f} nm."
        )

    wmin = w[wnmin > w][-1]
    wmax = w[wnmax < w][0]
    iwmin = w.tolist().index(wmin)
    iwmax = w.tolist().index(wmax)
    paths = [
        f"{path}/us76_u86_4-spectra-{w[i]}_{w[i+1]}.nc" for i in range(iwmin, iwmax)
    ]

    if len(paths) == 1:
        absorption_dataset = data.open_dataset(paths[0])
    else:
        datasets = [
            data.open_dataset(path).isel(w=slice(0, -1))  # strip endpoints
            for path in paths
        ]
        absorption_dataset = xr.concat(datasets, dim="w")

    return absorption_dataset


@parse_docs
@attrs.define(eq=False, slots=False)
class MolecularAtmosphere(AbstractHeterogeneousAtmosphere):
    """
    Molecular atmosphere scene element [``molecular``].

    .. admonition:: Class method constructors

       .. autosummary::

          afgl_1986
          ussa_1976
    """

    _thermoprops: xr.Dataset = documented(
        attrs.field(
            factory=us76.make_profile,
            validator=attrs.validators.instance_of(xr.Dataset),
            repr=summary_repr,
        ),
        doc="Thermophysical properties.",
        type="Dataset",
        default=":meth:`us76.make_profile() <eradiate.thermoprops.us76.make_profile>`",
    )

    _phase: PhaseFunction = documented(
        attrs.field(
            factory=lambda: RayleighPhaseFunction(),
            converter=phase_function_factory.convert,
            validator=attrs.validators.instance_of(PhaseFunction),
        ),
        doc="Phase function.",
        type=":class:`.PhaseFunction`",
        init_type=":class:`.PhaseFunction` or dict",
        default=":class:`RayleighPhaseFunction() <.RayleighPhaseFunction>`",
    )

    has_absorption: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
        ),
        doc="Absorption switch. If ``True``, the absorption coefficient is "
        "computed. Else, the absorption coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    has_scattering: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
        ),
        doc="Scattering switch. If ``True``, the scattering coefficient is "
        "computed. Else, the scattering coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    @has_absorption.validator
    @has_scattering.validator
    def _switch_validator(self, attribute, value):
        if not self.has_absorption and not self.has_scattering:
            raise ValueError(
                f"while validating {attribute.name}: at least one of 'has_absorption' "
                "and 'has_scattering' must be True"
            )

    absorption_dataset: xr.Dataset | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(
                converters.to_dataset(load_from_id=None)
            ),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(xr.Dataset)
            ),
            repr=summary_repr,
        ),
        doc="Absorption coefficient dataset. If ``None``, the absorption "
        "coefficient is set to zero.",
        type="Dataset",
        init_type="PathLike or Dataset",
        default="None",
    )

    _radprops_profile: RadProfile | None = attrs.field(
        default=None,
        validator=attrs.validators.optional(attrs.validators.instance_of(RadProfile)),
        init=False,
        repr=False,
    )

    @property
    def spectral_set(self) -> None | BinSet | WavelengthSet:
        if self.has_absorption:
            if eradiate.mode().is_mono:
                return WavelengthSet.from_absorption_dataset(
                    dataset=self.absorption_dataset
                )
            elif eradiate.mode().is_ckd:
                return BinSet.from_absorption_dataset(
                    dataset=self.absorption_dataset,
                    quad=Quad.gauss_legendre(16),  # TODO: PR#311 hack
                )
            else:
                raise NotImplementedError
        else:
            return None

    def update(self) -> None:
        # Inherit docstring

        self.phase.id = self.phase_id

        if self.thermoprops.title == "U.S. Standard Atmosphere 1976":
            # temporary fix
            if self.absorption_dataset is None and self.has_absorption:
                self.absorption_dataset = (
                    "spectra/absorption/us76_u86_4/us76_u86_4-spectra-18000_19000.nc"
                )

            self._radprops_profile = US76ApproxRadProfile(
                thermoprops=self.thermoprops,
                has_scattering=self.has_scattering,
                has_absorption=self.has_absorption,
                absorption_dataset=self.absorption_dataset,
            )
        elif "AFGL (1986)" in self.thermoprops.title:
            self._radprops_profile = AFGL1986RadProfile(
                thermoprops=self.thermoprops,
                has_scattering=self.has_scattering,
                has_absorption=self.has_absorption,
                absorption_dataset=self.absorption_dataset,
            )
        else:
            raise NotImplementedError("Unsupported thermophysical property data set.")

    # --------------------------------------------------------------------------
    #              Spatial extension and thermophysical properties
    # --------------------------------------------------------------------------

    @property
    def thermoprops(self) -> xr.Dataset:
        # Inherit docstring
        return self._thermoprops

    def eval_mfp(self, ctx: KernelContext) -> pint.Quantity:
        # Inherit docstring
        min_sigma_s = self.radprops_profile.eval_sigma_s(ctx.si).min()
        return np.divide(
            1.0,
            min_sigma_s,
            where=min_sigma_s != 0.0,
            out=np.array([np.inf]),
        )

    # --------------------------------------------------------------------------
    #                             Radiative properties
    # --------------------------------------------------------------------------

    @property
    def phase(self) -> PhaseFunction:
        # Inherit docstring
        return self._phase

    @property
    def radprops_profile(self) -> RadProfile:
        # Inherit docstring
        return self._radprops_profile

    def eval_albedo(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_albedo(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_t(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_sigma_t(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_a(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_sigma_a(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_s(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_sigma_s(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    # --------------------------------------------------------------------------
    #                             Kernel dictionary
    # --------------------------------------------------------------------------

    @property
    def _template_phase(self) -> dict:
        # Inherit docstring
        result, _ = traverse(self.phase)
        return result.data

    @property
    def _params_phase(self) -> dict:
        # Inherit docstring
        _, result = traverse(self.phase)
        return result.data

    # --------------------------------------------------------------------------
    #                               Constructors
    # --------------------------------------------------------------------------

    @classmethod
    def afgl_1986(
        cls,
        model: str = "us_standard",
        levels: pint.Quantity | None = None,
        concentrations: dict[str, str | pint.Quantity] | None = None,
        binset: str = "10nm",
        **kwargs: dict[str],
    ) -> MolecularAtmosphere:
        """
        Molecular atmosphere based on the AFGL (1986) atmospheric
        thermophysical properties profiles
        :cite:`Anderson1986AtmosphericConstituentProfiles` (CKD mode only).

        Parameters
        ----------
        model : {"us_standard", "tropical", "midlatitude_summer", \
            "midlatitude_winter", "subarctic_summer", "subarctic_winter"}, \
            optional, default: "us_standard"
            AFGL (1986) model identifier.

        levels : quantity
            Altitude levels.

        concentrations : dict
            Molecular concentrations as a ``{str: quantity}`` mapping.
            This dictionary is interpreted by :func:`pinttr.util.ensure_units`,
            which allows for passing units as strings.

        binset: str
            Wavelength bin set identifier. Either ``"10nm"`` or ``"1nm"``.

        **kwargs
            Keyword arguments passed to the :class:`.MolecularAtmosphere`
            constructor.

        Returns
        -------
        :class:`MolecularAtmosphere`
            AFGL (1986) molecular atmosphere.

        Notes
        -----
        :cite:`Anderson1986AtmosphericConstituentProfiles` defines six models,
        listed in the table below.

        .. list-table:: AFGL (1986) atmospheric thermophysical properties profiles models
           :widths: 2 4 4
           :header-rows: 1

           * - Model number
             - Model identifier
             - Model name
           * - 1
             - ``tropical``
             - Tropic (15N Annual Average)
           * - 2
             - ``midlatitude_summer``
             - Mid-Latitude Summer (45N July)
           * - 3
             - ``midlatitude_winter``
             - Mid-Latitude Winter (45N Jan)
           * - 4
             - ``subarctic_summer``
             - Sub-Arctic Summer (60N July)
           * - 5
             - ``subarctic_winter``
             - Sub-Arctic Winter (60N Jan)
           * - 6
             - ``us_standard``
             - U.S. Standard (1976)

        .. attention::

           The original altitude mesh specified by
           :cite:`Anderson1986AtmosphericConstituentProfiles` is a piecewise
           regular altitude mesh with an altitude step of 1 km from 0 to 25 km,
           2.5 km from 25 km to 50 km and 5 km from 50 km to 120 km.
           Since the Eradiate kernel only supports regular altitude mesh, the
           original atmospheric thermophysical properties profiles were
           interpolated on the regular altitude mesh with an altitude step of 1 km
           from 0 to 120 km.

        Although the altitude meshes of the interpolated
        :cite:`Anderson1986AtmosphericConstituentProfiles` profiles is fixed,
        this class lets you define a custom altitude mesh (regular or
        irregular).

        All six models include the following six absorbing molecular species:
        H2O, CO2, O3, N2O, CO, CH4 and O2.
        The concentrations of these species in the atmosphere is fixed by
        :cite:`Anderson1986AtmosphericConstituentProfiles`.
        However, this class allows you to rescale the concentrations of each
        individual molecular species to custom concentration values.
        Custom concentrations can be provided in different units.
        """
        if "absorption_dataset" in kwargs:
            raise TypeError(
                "Cannot pass 'absorption_dataset' keyword argument. The "
                "'afgl_1986' constructor sets the absorption dataset "
                "automatically."
            )

        thermoprops = afgl_1986.make_profile(model_id=model)

        # open the absorption dataset corresponding to 'binset' and 'model'
        path = "ckd/absorption"

        if binset == "10nm":
            absorption_dataset = f"{path}/10nm/afgl_1986-{model}-10nm-v3.nc"
        elif binset == "1nm":
            absorption_dataset = f"{path}/1nm/afgl_1986-{model}-1nm-v3.nc"
        else:
            raise ValueError(f"Invalid binset: {binset}")

        if concentrations is not None:
            factors = compute_scaling_factors(
                ds=thermoprops,
                concentration=pinttr.interpret_units(concentrations, ureg=ureg),
            )
            thermoprops = rescale_concentration(ds=thermoprops, factors=factors)

        if levels is not None:
            thermoprops = interpolate(
                ds=thermoprops,
                z_level=levels,
                conserve_columns=True,
            )

        return cls(
            thermoprops=thermoprops,
            absorption_dataset=absorption_dataset,
            **kwargs,
        )

    @classmethod
    def ussa_1976(
        cls,
        levels: pint.Quantity | None = None,
        concentrations: dict[str, pint.Quantity] | None = None,
        wavelength_range: pint.Quantity | None = None,
        **kwargs: dict[str, t.Any],
    ) -> MolecularAtmosphere:
        """
        Molecular atmosphere based on the US Standard Atmosphere (1976) model
        :cite:`NASA1976USStandardAtmosphere` (monochromatic mode only).

        Parameters
        ----------
        levels : quantity, optional
            Altitude levels. If ``None``, defaults to [0, 1, ..., 99, 100] km.

        concentrations : dict
            Molecules concentrations as a ``{str: quantity}`` mapping.

        wavelength_range: quantity
            Wavelength range wherein the atmosphere radiative properties are
            expected to be computed.
            This information is used to select the monochromatic absorption
            dataset(s) to open.
            If None, defaults to [545.0, 555.0] nm.

        **kwargs
            Keyword arguments passed to the :class:`.MolecularAtmosphere`
            constructor.

        Returns
        -------
        :class:`.MolecularAtmosphere`
            U.S. Standard Atmosphere (1976) molecular atmosphere object.
        """
        if "absorption_dataset" in kwargs:
            raise TypeError(
                "Cannot pass 'absorption_dataset' keyword argument. The "
                "'ussa_1976' constructor sets the absorption dataset "
                "automatically."
            )

        # open the absorption dataset corresponding to 'wavelength_range
        if wavelength_range is None:
            wavelength_range = DEFAULT_WAVELENGTH_RANGE

        absorption_dataset = open_us76_dataset(wavelength_range)

        if levels is None:
            levels = np.linspace(0, 100, 101) * ureg.km

        ds = us76.make_profile(levels=levels)

        if concentrations is not None:
            factors = compute_scaling_factors(ds=ds, concentration=concentrations)
            ds = rescale_concentration(ds=ds, factors=factors)

        return cls(
            thermoprops=ds,
            absorption_dataset=absorption_dataset,
            **kwargs,
        )



from . import absorption as absorption
from . import rayleigh as rayleigh
from ._afgl1986 import AFGL1986RadProfile as AFGL1986RadProfile
from ._core import RadProfile as RadProfile
from ._core import ZGrid as ZGrid
from ._core import rad_profile_factory as rad_profile_factory
from ._us76_approx import US76ApproxRadProfile as US76ApproxRadProfile



import numpy as np
import pytest

import eradiate
from eradiate.experiments import AtmosphereExperiment
from eradiate.pipelines import ApplySpectralResponseFunction
from eradiate.scenes.measure import MultiDistantMeasure
from eradiate.units import unit_registry as ureg


@pytest.mark.parametrize(
    "mode_id, srf",
    (
        ("mono", {"type": "multi_delta", "wavelengths": 550.0 * ureg.nm}),
        (
            "ckd",
            {
                "type": "interpolated",
                "wavelengths": [550, 560] * ureg.nm,
                "values": [1.0, 1.0] * ureg.dimensionless,
            },
        ),
    ),
    ids=("mono", "ckd_multiple_bins"),
)
def test_apply_spectral_response_function_transform(mode_id, srf):
    """
    Unit tests for ApplySpectralResponseFunction.transform().
    """
    # Prepare basic data
    eradiate.set_mode(mode_id)

    exp = AtmosphereExperiment(
        atmosphere=None,
        measures=MultiDistantMeasure.hplane(
            id="measure",
            zeniths=[-60, -45, 0, 45, 60],
            azimuth=0.0,
            spp=1,
            srf=srf,
        ),
    )
    measure = exp.measures[0]
    exp.process()

    # Apply first steps of post-processing
    pipeline = exp.pipeline(measure)
    values = pipeline.transform(measure.mi_results, stop_after="add_viewing_angles")

    # Apply tested pipeline step
    step = ApplySpectralResponseFunction(measure=measure, vars=["radiance"])

    if eradiate.mode().is_mono:
        # In mono modes, the dataset produced by previous steps is missing
        # bin data required for the step to run successfully
        with pytest.raises(ValueError):
            step.transform(values)
        return

    # In binned modes, computation goes through
    result = step.transform(values)

    # The step adds a SRF-weighted variable
    assert "radiance_srf" in result.data_vars
    assert np.all(result.radiance_srf > 0.0)



import itertools
import logging
import typing as t
from collections import OrderedDict

import attrs
import numpy as np
import xarray as xr

from ._core import PipelineStep
from ..attrs import documented, parse_docs
from ..scenes.measure import Measure
from ..spectral.ckd import BinSet
from ..units import symbol
from ..units import unit_context_config as ucc
from ..units import unit_context_kernel as uck

logger = logging.getLogger(__name__)


@parse_docs
@attrs.define
class AggregateCKDQuad(PipelineStep):
    """
    Compute CKD quadrature.

    In CKD modes, this pipeline step aggregates spectral data and
    evaluates the selected quadrature rule. The following updates to the input
    data are expected:

    * the ``index`` dimension is dropped;
    * the ``bin`` dimension is renamed ``w``;
    * the ``bin`` coordinate persists and is indexed by ``w``;
    * a ``w`` coordinate is created and contains the central wavelength of each
      bin;
    * a ``bin_wmin`` (resp. ``bin_wmax``) coordinate is created and contains the
      lower (resp. upper) spectral bound of each bin;
    * the dataset is reordered by ascending ``w`` values.

    Notes
    -----
    * The ``spp`` variable is averaged on the ``index`` dimension.
    * In non-CKD modes, this step is a no-op.
    """

    measure: Measure = documented(
        attrs.field(
            validator=attrs.validators.instance_of(Measure),
            repr=lambda self: f"{self.__class__.__name__}(id='{self.id}', ...)",
        ),
        doc="Measure from which the processed data was obtained.",
        type=":class:`.Measure`",
    )

    binset: BinSet = documented(
        attrs.field(validator=attrs.validators.instance_of(BinSet)),
        doc="Bin set.",
        type=":class:`.BinSet`",
    )

    var: str = documented(
        attrs.field(default="img", validator=attrs.validators.instance_of(str)),
        doc="Name of the variable for which CKD quadrature computation "
        "is to be performed.",
        type="str",
        default='"img"',
    )

    def transform(self, x: t.Any) -> t.Any:
        logger.debug("aggregate_ckd_quad: begin")
        # Compute quadrature spectrum-indexed variables and turn spp
        # into a per-bin average
        var = self.var

        # Get dimensions of current variable
        img = x.data_vars[var]
        dims = OrderedDict((y, len(img.coords[y])) for y in img.dims)

        if "bin" not in dims:
            raise ValueError(f"variable '{var}' is missing dimension 'bin'")

        if "index" not in dims:
            raise ValueError(f"variable '{var}' is missing dimension 'index'")

        # Init storage
        del dims["bin"]
        del dims["index"]

        # Deduplicate bin list preserving order
        bin_ids = list(OrderedDict.fromkeys(x.bin.to_index()))
        bin_ids = [str(int(_)) for _ in x.bin.values]
        logger.debug(f"{bin_ids=}")

        n_bins = len(bin_ids)

        # Collect wavelengths associated with each bin
        wavelength_units = ucc.get("wavelength")
        wavelengths = np.array(x.bin.values, dtype=float)
        bin_wmins = []
        bin_wmaxs = []
        for bin in self.binset.bins:
            wmin, wmax = bin.wmin, bin.wmax
            bin_wmins.append(wmin.m_as(wavelength_units))
            bin_wmaxs.append(wmax.m_as(wavelength_units))

        result = x

        aggregated = xr.DataArray(
            np.zeros([n_bins] + list(dims.values())),
            coords={"bin": img.bin, **{dim: img.coords[dim] for dim in dims}},
        )

        # TODO: PR#311 hack (next 3 lines)
        _bins = self.binset.bins
        _bins_wcenters = [b.wcenter.m_as(wavelength_units) for b in _bins]
        _bins_wcenters_str = [str(int(w)) for w in _bins_wcenters]

        # For each bin and each pixel, compute quadrature and store the result
        for i_bin, bin in enumerate(bin_ids):
            # TODO: PR#311 hack (next 2 lines)
            _bin_index = _bins_wcenters_str.index(bin)
            _bin = _bins[_bin_index]

            values_at_nodes = img.sel(bin=bin).values

            # Rationale: Avoid using xarray's indexing in this loop for
            # performance reasons (wrong data indexing method will result in
            # 10x+ speed reduction)
            for indexes in itertools.product(*[list(range(n)) for n in dims.values()]):
                aggregated.values[(i_bin, *indexes)] = _bin.quad.integrate(
                    values_at_nodes[(slice(None), *indexes)],
                    interval=np.array([0.0, 1.0]),
                )

        result = result.assign({var: aggregated})
        result[var].attrs = x[var].attrs

        # Average the 'spp' variable over the 'index' dimension
        with xr.set_options(keep_attrs=True):
            result["spp"] = x.spp.mean(dim="index")

        # Add spectral coordinates
        result = result.assign_coords(
            {
                "w": (
                    "bin",
                    wavelengths,
                    {
                        "standard_name": "wavelength",
                        "long_name": "wavelength",
                        "units": symbol(wavelength_units),
                    },
                ),
                "bin_wmin": (
                    "bin",
                    bin_wmins,
                    {
                        "standard_name": "bin_wmin",
                        "long_name": "spectral bin lower bound",
                        "units": symbol(wavelength_units),
                    },
                ),
                "bin_wmax": (
                    "bin",
                    bin_wmaxs,
                    {
                        "standard_name": "bin_wmax",
                        "long_name": "spectral bin upper bound",
                        "units": symbol(wavelength_units),
                    },
                ),
            }
        )

        # Swap the 'bin' and 'w' dimensions
        result = result.swap_dims({"bin": "w"})

        # Remove the 'index' dimension
        result = result.drop_dims("index")

        # Reorder by ascending "w"
        result = result.sortby("w")

        logger.debug("aggregate_ckd_quad: end")

        return result


@parse_docs
@attrs.define
class AggregateRadiosity(PipelineStep):
    """
    Aggregate flux density field.
    """

    sector_radiosity_var: str = documented(
        attrs.field(
            default="sector_radiosity", validator=attrs.validators.instance_of(str)
        ),
        doc="Name of the variable containing radiosity values for the "
        "hemisphere sector corresponding to each film pixel. This quantity is "
        "expressed in flux units (typically W/m²) and, when summed over the "
        "entire film, aggregates into a radiosity.",
        type="str",
        default='"sector_radiosity"',
    )

    radiosity_var: str = documented(
        attrs.field(default="radiosity", validator=attrs.validators.instance_of(str)),
        doc="Name of the variable storing the computed radiosity "
        "(leaving flux) value.",
        type="str",
        default='"radiosity"',
    )

    def transform(self, x: t.Any) -> t.Any:
        result = x.copy(deep=False)

        result[self.radiosity_var] = result[self.sector_radiosity_var].sum(
            dim=("x_index", "y_index")
        )
        result[self.radiosity_var].attrs = {
            "standard_name": "toa_outgoing_flux_density_per_unit_wavelength",
            "long_name": "top-of-atmosphere outgoing spectral flux density",
            "units": symbol(uck.get("irradiance")),
        }

        return result



"""
US Standard Atmosphere 1976 thermophysical model, according to
:cite:`NASA1976USStandardAtmosphere`.
"""

from datetime import datetime

import numpy as np
import numpy.ma as ma
import xarray as xr
from scipy.integrate import cumtrapz
from scipy.interpolate import interp1d

from eradiate import __version__

from ..units import unit_registry as ureg

# ------------------------------------------------------------------------------
#
# Atmospheric vertical profile data set generator
#
# ------------------------------------------------------------------------------


@ureg.wraps(ret=None, args="m", strict=False)
def make_profile(levels=ureg.Quantity(np.linspace(0.0, 1e5, 51), "m")):
    r"""
    Makes an atmosphere vertical profile based on the US Standard Atmosphere
    1976 thermophysical model.

    Parameters
    ----------
    levels : array-like, default: np.linspace(0, 1e5, 51) * ureg.m
        Levels altitudes [m]. The values must be sorted by increasing order.
        Valid range: 0 to 1e6 m.

    Returns
    -------
    Dataset
        Data set holding the values of the pressure, temperature,
        total number density and number densities of the individual
        gas species in each layer.

    Notes
    -----
    The pressure, temperature and number densities given in each layer of
    the altitude mesh are computed at the altitude of the layers centers.
    In other words, the layer's middle is taken as the altitude
    representative of the whole layer. For example, in a layer with lower
    and upper altitudes of 1000 and 2000 m, the thermophysical variables
    are computed at the altitude of 1500 m.
    """

    if np.any(levels > 1e6) or np.any(levels < 0.0):
        raise ValueError("Levels altitudes must be in [0, 1e6] m.")

    z_layer = (levels[:-1] + levels[1:]) / 2

    # create the US76 data set
    ds = create(
        ureg.Quantity(z_layer, "m"),
        variables=["p", "t", "n", "n_tot"],
    )

    # derive atmospheric thermophysical properties profile data set
    thermoprops_ds = (
        xr.Dataset(
            data_vars={
                "p": ds.p,
                "t": ds.t,
                "n": ds.n_tot,
                "mr": ds.n / ds.n_tot,
            }
        )
        .rename_dims({"z": "z_layer"})
        .drop_vars("z")
    )
    thermoprops_ds.data_vars["mr"].attrs.update(
        dict(standard_name="mixing_ratio", long_name="mixing ratio", units="")
    )
    thermoprops_ds.coords["z_layer"] = (
        "z_layer",
        z_layer,
        dict(
            standard_name="layer_altitude",
            long_name="layer altitude",
            units="m",
        ),
    )
    thermoprops_ds.coords["z_level"] = (
        "z_level",
        levels,
        dict(
            standard_name="level_altitude",
            long_name="level altitude",
            units="m",
        ),
    )
    thermoprops_ds.attrs = dict(
        convention="CF-1.8",
        title="U.S. Standard Atmosphere 1976",
        history=f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} - "
        f"data creation - eradiate.scenes.atmosphere.us76.make_profile",
        source=f"eradiate, version {__version__}",
        references="U.S. Standard Atmosphere, 1976, NASA-TM-X-74335, "
        "NOAA-S/T-76-1562",
    )

    return thermoprops_ds


# ------------------------------------------------------------------------------
#
# Constants.
# As much as possible, constants names are chosen to be as close as possible to
# the notations used in :cite:`NASA1976USStandardAtmosphere`.
#
# ------------------------------------------------------------------------------

# Boltzmann constant
K = 1.380622e-23  # [J/K]

# Molar masses of the individual species
M = {
    "N2": 0.0280134,
    "O2": 0.0319988,
    "Ar": 0.039948,
    "CO2": 0.04400995,
    "Ne": 0.020183,
    "He": 0.0040026,
    "Kr": 0.08380,
    "Xe": 0.13130,
    "CH4": 0.01604303,
    "H2": 0.00201594,
    "O": 0.01599939,
    "H": 0.00100797,
}  # [kg/mol]

# Sea level mean (mixture) molar mass
M0 = 0.028964425278793997  # [kg/mol]

# Avogadro number
NA = 6.022169e23  # [mol^-1]

# Universal gas constant
R = 8.31432  # [J/(mol*K)]

# Sea level volume fractions of the gas species present below 86 km
F = {
    "N2": 0.78084,
    "O2": 0.209476,
    "Ar": 0.00934,
    "CO2": 0.000314,
    "Ne": 0.00001818,
    "He": 0.00000524,
    "Kr": 0.00000114,
    "Xe": 0.000000087,
    "CH4": 0.000002,
    "H2": 0.0000005,
}  # [dimensionless]

# Sea level gravity
G0 = 9.80665  # [m/s^2]

# Geopotential altitudes of the layers' boundaries (below 86 km)
H = [
    0.0,
    11e3,
    20e3,
    32e3,
    47e3,
    51e3,
    71e3,
    84852.05,
]  # [m]

# Temperature gradients in the seven layers (below 86 km)
LK = [
    -0.0065,
    0.0,
    0.0010,
    0.0028,
    0.0,
    -0.0028,
    -0.0020,
]  # [K/m]

# Pressure at sea level
P0 = 101325.0  # [Pa]

# Effective Earth radius
R0 = 6.356766e6  # [m]

# Temperature at sea level
T0 = 288.15  # [K]
S = 110.4  # [K]
BETA = 1.458e6  # [kg/(m*s*K^1/2)]
GAMMA = 1.40  # [dimensionless]
SIGMA = 3.65e-10  # [m]

# Thermal diffusion constants of the individual species present above 86 km
ALPHA = {
    "N2": 0.0,  # [dimensionless]
    "O": 0.0,  # [dimensionless]
    "O2": 0.0,  # [dimensionless]
    "Ar": 0.0,  # [dimensionless]
    "He": -0.4,  # [dimensionless]
    "H": -0.25,  # [dimensionless]
}
A = {
    "N2": None,
    "O": 6.986e20,  # [m^-1*s^-1]
    "O2": 4.863e20,  # [m^-1*s^-1]
    "Ar": 4.487e20,  # [m^-1*s^-1]
    "He": 1.7e21,  # [m^-1*s^-1]
    "H": 3.305e21,  # [m^-1*s^-1]
}
B = {
    "N2": None,
    "O": 0.75,  # [dimensionless]
    "O2": 0.75,  # [dimensionless]
    "Ar": 0.87,  # [dimensionless]
    "He": 0.691,  # [dimensionless]
    "H": 0.5,  # [dimensionless]
}

# Eddy diffusion coefficients
K_7 = 1.2e2  # [m^2/s]
K_10 = 0.0  # [m^2/s]

# Vertical transport constants of the individual species present above 86 km
Q1 = {
    "O": -5.809644e-4,  # [km^-3]
    "O2": 1.366212e-4,  # [km^-3]
    "Ar": 9.434079e-5,  # [km^-3]
    "He": -2.457369e-4,  # [km^-3]
}
Q2 = {
    "O": -3.416248e-3,  # [km^-3], /!\ above 97 km, Q2 = 0.
    "O2": 0.0,  # [km^-3]
    "Ar": 0.0,  # [km^-3]
    "He": 0.0,  # [km^-3]
}
U1 = {
    "O": 56.90311,  # [km]
    "O2": 86.0,  # [km]
    "Ar": 86.0,  # [km]
    "He": 86.0,  # [km]
}
U2 = {"O": 97.0, "O2": None, "Ar": None, "He": None}  # [km]
W1 = {
    "O": 2.706240e-5,  # [km^-3]
    "O2": 8.333333e-5,  # [km^-3]
    "Ar": 8.333333e-5,  # [km^-3]
    "He": 6.666667e-4,  # [km^-3]
}
W2 = {"O": 5.008765e-4, "O2": None, "Ar": None, "He": None}  # [km^-3]

# Altitudes of the levels delimiting 5 layers above 86 km
Z7 = 86.0  # [km]
Z8 = 91.0  # [km]
Z9 = 110.0  # [km]
Z10 = 120.0  # [km]
Z11 = 500.0  # [km]
Z12 = 1000.0  # [km]

# Temperature at the different levels above 86 km
T7 = 186.8673  # [K]
T9 = 240.0  # [K]
T10 = 360.0  # [K]
T11 = 999.2356  # [K]
TINF = 1000.0  # [K]
LAMBDA = 0.01875  # [km^-1]

# Temperature gradients
LK7 = 0.0  # [K/km]
LK9 = 12.0  # [K/km]

# Molecular nitrogen at altitude = Z7
N2_7 = 1.129794e20  # [m^-3]

# Atomic oxygen at altitude = Z7
O_7 = 8.6e16  # [m^-3]

# Molecular oxygen at altitude = Z7
O2_7 = 3.030898e19  # [m^-3]

# Argon at altitude = Z7
AR_7 = 1.351400e18  # [m^-3]

# Helium at altitude = Z7 (assumes typo at page 13)
HE_7 = 7.5817e14  # [m^-3]

# Hydrogen at altitude = Z7
H_11 = 8.0e10  # [m^-3]

# Vertical flux
PHI = 7.2e11  # [m^-2 * s^-1]

# List of all gas species
SPECIES = ["N2", "O2", "Ar", "CO2", "Ne", "He", "Kr", "Xe", "CH4", "H2", "O", "H"]

# List of variables computed by the model
VARIABLES = [
    "t",
    "p",
    "n",
    "n_tot",
    "rho",
    "mv",
    "hp",
    "v",
    "mfp",
    "f",
    "cs",
    "mu",
    "nu",
    "kt",
]

# Variables standard names with respect to the Climate and Forecast (CF)
# convention
STANDARD_NAME = {
    "t": "air_temperature",
    "p": "air_pressure",
    "n": "number_density",
    "n_tot": "air_number_density",
    "rho": "air_density",
    "mv": "molar_volume",
    "hp": "pressure_scale_height",
    "v": "mean_air_particles_speed",
    "mfp": "mean_free_path",
    "f": "mean_collision_frequency",
    "cs": "speed_of_sound_in_air",
    "mu": "air_dynamic_viscosity",
    "nu": "air_kinematic_viscosity",
    "kt": "air_thermal_conductivity_coefficient",
    "z": "altitude",
    "h": "geopotential_height",
}

# Units of relevant quantities
UNITS = {
    "t": "K",
    "p": "Pa",
    "n": "m^-3",
    "n_tot": "m^-3",
    "rho": "kg/m^3",
    "mv": "m^3",
    "hp": "m",
    "v": "m/s",
    "mfp": "m",
    "f": "s^-1",
    "cs": "m/s",
    "mu": "kg/(m*s)",
    "nu": "m^2/s",
    "kt": "J/m",
    "z": "m",
    "h": "m",
    "species": "",
}

# Variables dimensions
DIMS = {
    "t": "z",
    "p": "z",
    "n": ("species", "z"),
    "n_tot": "z",
    "rho": "z",
    "mv": "z",
    "hp": "z",
    "v": "z",
    "mfp": "z",
    "f": "z",
    "cs": "z",
    "mu": "z",
    "nu": "z",
    "kt": "z",
}


# ------------------------------------------------------------------------------
#
# Computational functions.
# The U.S. Standard Atmosphere 1976 model divides the atmosphere into two
# altitude regions:
#   1. the low-altitude region, from 0 to 86 kilometers
#   2. the high-altitude region, from 86 to 1000 kilometers.
# The majority of computational functions hereafter are specialized for one or
# the other altitude region and is valid only in that altitude region, not in
# the other.
#
# ------------------------------------------------------------------------------


@ureg.wraps(ret=None, args=("m", None), strict=False)
def create(z, variables=None):
    """
    Creates a US Standard Atmosphere 1976 data set using specified altitudes
    values.

    Parameters
    ----------
    z : array-like
        1-D array with altitude values [m].

    variables : list, optional
        Names of the variables to compute.

    Returns
    -------
    Dataset
        Data set holding the values of the different atmospheric variables.

    Warnings
    --------
    The returned U.S. Standard Atmosphere 1976 data set is not an
    atmospheric vertical profile data set. See
    :func:`eradiate.thermoprops.us76.make_profile`
    if you are interested in generating an atmospheric vertical profile
    based on the U.S. Standard Atmosphere 1976 model.
    """

    if np.any(z < 0.0):
        raise ValueError("altitude values must be greater than or equal to " "zero")

    if np.any(z > 1000000.0):
        raise ValueError("altitude values must be less then or equal to 1e6 m")

    if variables is None:
        variables = VARIABLES
    else:
        for var in variables:
            if var not in VARIABLES:
                raise ValueError(var, " is not a valid variable name")

    # initialise data set
    ds = init_data_set(ureg.Quantity(z, "m"))

    # compute the model in the low-altitude region
    compute_low_altitude(ds, ds.coords["z"] <= 86000.0, inplace=True)

    # compute the model in the high-altitude region
    compute_high_altitude(ds, ds.coords["z"] > 86000.0, inplace=True)

    # replace all np.nan with 0. in number densities values
    n = ds.n.values
    n[np.isnan(n)] = 0.0
    ds.n.values = n

    # list names of variables to drop from the data set
    names = []
    for var in ds.data_vars:
        if var not in variables:
            names.append(var)

    return ds.drop_vars(names)


def compute_low_altitude(data_set, mask=None, inplace=False):
    r"""
    Computes the US Standard Atmosphere 1976 in the low-altitude region.

    Parameters
    ----------
    data_set : Dataset
        Data set to compute.

    mask : DataArray, optional
        Mask to select the region of the data set to compute.
        By default, the mask selects the entire data set.

    inplace : bool, default: False
        If true, modifies ``data_set`` in place, else returns a copy of
        ``data_set``.
        Default: False.

    Returns
    -------
    Dataset or None
        If ``inplace`` is True, returns nothing, else returns a copy of
        ``data_set``.
        the values of the computed variables.
    """

    if mask is None:
        mask = xr.full_like(data_set.coords["z"], True, dtype=bool)

    if inplace:
        ds = data_set
    else:
        ds = data_set.copy(deep=True)

    altitudes = ds.coords["z"][mask]
    z = altitudes.values

    # compute levels temperature and pressure values
    tb, pb = compute_levels_temperature_and_pressure_low_altitude()

    # compute geopotential height, temperature and pressure
    h = to_geopotential_height(z)
    t = compute_temperature_low_altitude(h, tb)
    p = compute_pressure_low_altitude(h, pb, tb)

    # compute the auxiliary atmospheric variables
    n_tot = NA * p / (R * t)
    rho = p * M0 / (R * t)
    g = compute_gravity(z)
    mu = BETA * np.power(t, 1.5) / (t + S)

    # assign data set with computed values
    ds["t"].loc[dict(z=altitudes)] = t
    ds["p"].loc[dict(z=altitudes)] = p
    ds["n_tot"].loc[dict(z=altitudes)] = n_tot

    species = ["N2", "O2", "Ar", "CO2", "Ne", "He", "Kr", "Xe", "CH4", "H2"]
    for i, s in enumerate(SPECIES):
        if s in species:
            ds["n"][i].loc[dict(z=altitudes)] = F[s] * n_tot

    ds["rho"].loc[dict(z=altitudes)] = rho
    ds["mv"].loc[dict(z=altitudes)] = NA / n_tot
    ds["hp"].loc[dict(z=altitudes)] = R * t / (g * M0)
    ds["v"].loc[dict(z=altitudes)] = np.sqrt(8.0 * R * t / (np.pi * M0))
    ds["mfp"].loc[dict(z=altitudes)] = np.sqrt(2.0) / (
        2.0 * np.pi * np.power(SIGMA, 2.0) * n_tot
    )
    ds["f"].loc[dict(z=altitudes)] = (
        4.0
        * NA
        * np.power(SIGMA, 2.0)
        * np.sqrt(np.pi * np.power(p, 2.0) / (R * M0 * t))
    )
    ds["cs"].loc[dict(z=altitudes)] = np.sqrt(GAMMA * R * t / M0)
    ds["mu"].loc[dict(z=altitudes)] = mu
    ds["nu"].loc[dict(z=altitudes)] = mu / rho
    ds["kt"].loc[dict(z=altitudes)] = (
        2.64638e-3 * np.power(t, 1.5) / (t + 245.4 * np.power(10.0, -12.0 / t))
    )

    if not inplace:
        return ds


def compute_high_altitude(data_set, mask=None, inplace=False):
    r"""
    Computes the US Standard Atmosphere 1976 in the high-altitude region.

    Parameters
    ----------
    data_set : Dataset
        Data set to compute.

    mask : DataArray, optional
        Mask to select the region of the data set to compute.
        By default, the mask selects the entire data set.

    inplace : bool, default: False
        If true, modifies ``data_set`` in place, else returns a copy of
        ``data_set``.
        Default: False.

    Returns
    -------
    Dataset or None
        If ``inplace`` is True, returns nothing, else returns a copy of
        ``data_set``.
    """

    if mask is None:
        mask = xr.full_like(data_set.coords["z"], True, dtype=bool)

    if inplace:
        ds = data_set
    else:
        ds = data_set.copy(deep=True)

    altitudes = ds.coords["z"][mask]
    if len(altitudes) == 0:
        return ds

    z = ureg.Quantity(altitudes.values, "m")
    n = compute_number_densities_high_altitude(z)
    species = ["N2", "O", "O2", "Ar", "He", "H"]
    ni = np.array([n[s] for s in species])
    n_tot = np.sum(ni, axis=0)
    fi = ni / n_tot[np.newaxis, :]
    mi = np.array([M[s] for s in species])
    m = np.sum(fi * mi[:, np.newaxis], axis=0)
    t = compute_temperature_high_altitude(z)
    p = K * n_tot * t
    rho = np.sum(ni * mi[:, np.newaxis], axis=0) / NA
    g = compute_gravity(z)

    # assign data set with computed values
    ds["t"].loc[dict(z=altitudes)] = t
    ds["p"].loc[dict(z=altitudes)] = p
    ds["n_tot"].loc[dict(z=altitudes)] = n_tot

    for i, s in enumerate(SPECIES):
        if s in species:
            ds["n"][i].loc[dict(z=altitudes)] = n[s]

    ds["rho"].loc[dict(z=altitudes)] = rho
    ds["mv"].loc[dict(z=altitudes)] = NA / n_tot
    ds["hp"].loc[dict(z=altitudes)] = R * t / (g * m)
    ds["v"].loc[dict(z=altitudes)] = np.sqrt(8.0 * R * t / (np.pi * m))
    ds["mfp"].loc[dict(z=altitudes)] = np.sqrt(2.0) / (
        2.0 * np.pi * np.power(SIGMA, 2.0) * n_tot
    )
    ds["f"].loc[dict(z=altitudes)] = (
        4.0
        * NA
        * np.power(SIGMA, 2.0)
        * np.sqrt(np.pi * np.power(p, 2.0) / (R * m * t))
    )

    if not inplace:
        return ds


@ureg.wraps(ret=None, args="m", strict=False)
def init_data_set(z):
    r"""
    Initializes the data set.

    Parameters
    ----------
    z : array
        Altitudes values [m]

    Returns
    -------
    Dataset
        Initialised data set.
    """
    data_vars = {}
    for var in VARIABLES:
        if var != "n":
            try:
                data_vars[var] = (
                    DIMS[var],
                    np.full(z.shape, np.nan),
                    {"units": UNITS[var], "standard_name": STANDARD_NAME[var]},
                )
            except KeyError:
                data_vars[var] = (
                    DIMS[var],
                    np.full(z.shape, np.nan),
                    {"units": UNITS[var], "standard_name": STANDARD_NAME[var]},
                )
        else:
            data_vars[var] = (
                DIMS[var],
                np.full((len(SPECIES), len(z)), np.nan),
                {"units": UNITS[var], "standard_name": STANDARD_NAME["n"]},
            )

    coords = {"z": ("z", z, {"units": UNITS["z"]}), "species": ("species", SPECIES)}

    # TODO: set function name in history field dynamically
    attrs = {
        "convention": "CF-1.8",
        "title": "U.S. Standard Atmosphere 1976",
        "history": f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} - data creation - "
        f"eradiate.scenes.atmosphere.us76.create",
        "source": f"eradiate, version {__version__}",
        "references": "U.S. Standard Atmosphere, 1976, NASA-TM-X-74335, NOAA-S/T-76-1562",
    }

    return xr.Dataset(data_vars, coords, attrs)


def compute_levels_temperature_and_pressure_low_altitude():
    r"""
    Computes the temperature and the pressure values at the 8 levels
    of the low-altitude model.

    Returns
    -------
    array
         Levels temperature values [K].

    array
        Levels pressure values [Pa].
    """
    tb = [T0]
    pb = [P0]
    for i in range(1, 8):
        t_next = tb[i - 1] + LK[i - 1] * (H[i] - H[i - 1])
        tb.append(t_next)
        if LK[i - 1] == 0:
            p_next = compute_pressure_low_altitude_zero_gradient(
                H[i], H[i - 1], pb[i - 1], tb[i - 1]
            )
        else:
            p_next = compute_pressure_low_altitude_non_zero_gradient(
                H[i], H[i - 1], pb[i - 1], tb[i - 1], LK[i - 1]
            )
        pb.append(p_next)
    return tb, pb


@ureg.wraps(ret=None, args="km", strict=False)
def compute_number_densities_high_altitude(altitudes):
    r"""
    Computes the number density of the individual species in the
    high-altitude region.

    .. note::
        A uniform altitude grid is generated and used for the computation of the
        integral as well as for the computation of the number densities of the
        individual species. This gridded data is then interpolated at the query
        ``altitudes`` using a linear interpolation scheme in logarithmic space.

    Parameters
    ----------
    altitudes : array-like
        Altitude value(s) [km].

    Returns
    -------
    ndarray
        Number densities of the individual species and total number density at
        the given altitudes [m^-3].
        The number densities of the individual species are stored in a single
        2-D array where the first dimension is the gas species and the second
        dimension is the altitude. The species are in the following order:

        =====  ======
        Row    Species
        =====  ======
         0       N2
         1       O
         2       O2
         3       Ar
         4       He
         5       H
        =====  ======
    """

    # altitude grid
    grid = np.concatenate(
        (
            np.linspace(start=Z7, stop=150.0, num=640, endpoint=False),
            np.geomspace(start=150.0, stop=Z12, num=100, endpoint=True),
        )
    )  # [km]

    # pre-computed variables
    m = compute_mean_molar_mass_high_altitude(grid)  # [kg/mol]
    g = compute_gravity(ureg.Quantity(grid, "km"))  # [m / s^2]
    t = compute_temperature_high_altitude(grid)  # [K]
    dt_dz = compute_temperature_gradient_high_altitude(grid)  # [K/m]
    below_115 = grid < 115.0
    k = eddy_diffusion_coefficient(grid[below_115])  # [m^2/s]

    n_grid = {}

    # molecular nitrogen
    y = m * g / (R * t)  # [m^-1]
    n_grid["N2"] = (
        N2_7 * (T7 / t) * np.exp(-cumtrapz(y, 1e3 * grid, initial=0.0))
    )  # the factor 1000 is to convert km to m

    # atomic oxygen
    d = thermal_diffusion_coefficient(
        background=n_grid["N2"][below_115], temperature=t[below_115], a=A["O"], b=B["O"]
    )
    y = thermal_diffusion_term_atomic_oxygen(
        grid, g, t, dt_dz, d, k
    ) + velocity_term_atomic_oxygen(grid)
    n_grid["O"] = O_7 * (T7 / t) * np.exp(-cumtrapz(y, 1e3 * grid, initial=0.0))

    # molecular oxygen
    d = thermal_diffusion_coefficient(
        background=n_grid["N2"][below_115],
        temperature=t[below_115],
        a=A["O2"],
        b=B["O2"],
    )
    y = thermal_diffusion_term("O2", grid, g, t, dt_dz, m, d, k) + velocity_term(
        "O2", grid
    )
    n_grid["O2"] = O2_7 * (T7 / t) * np.exp(-cumtrapz(y, 1e3 * grid, initial=0.0))

    # argon
    background = (
        n_grid["N2"][below_115] + n_grid["O"][below_115] + n_grid["O2"][below_115]
    )
    d = thermal_diffusion_coefficient(
        background=background, temperature=t[below_115], a=A["Ar"], b=B["Ar"]
    )
    y = thermal_diffusion_term("Ar", grid, g, t, dt_dz, m, d, k) + velocity_term(
        "Ar", grid
    )
    n_grid["Ar"] = AR_7 * (T7 / t) * np.exp(-cumtrapz(y, 1e3 * grid, initial=0.0))

    # helium
    background = (
        n_grid["N2"][below_115] + n_grid["O"][below_115] + n_grid["O2"][below_115]
    )
    d = thermal_diffusion_coefficient(
        background=background, temperature=t[below_115], a=A["He"], b=B["He"]
    )
    y = thermal_diffusion_term("He", grid, g, t, dt_dz, m, d, k) + velocity_term(
        "He", grid
    )
    n_grid["He"] = HE_7 * (T7 / t) * np.exp(-cumtrapz(y, 1e3 * grid, initial=0.0))

    # hydrogen

    # below 500 km
    mask = (grid >= 150.0) & (grid <= 500.0)
    background = (
        n_grid["N2"][mask]
        + n_grid["O"][mask]
        + n_grid["O2"][mask]
        + n_grid["Ar"][mask]
        + n_grid["He"][mask]
    )
    d = thermal_diffusion_coefficient(background, t[mask], A["H"], B["H"])
    alpha = ALPHA["H"]
    _tau = tau_function(grid[mask], below_500=True)
    y = (PHI / d) * np.power(t[mask] / T11, 1 + alpha) * np.exp(_tau)
    integral_values = cumtrapz(y[::-1], 1e3 * grid[mask][::-1], initial=0.0)
    integral_values = integral_values[::-1]
    n_below_500 = (
        (H_11 - integral_values) * np.power(T11 / t[mask], 1 + alpha) * np.exp(-_tau)
    )

    # above 500 km
    _tau = tau_function(grid[grid > 500.0], below_500=False)
    n_above_500 = H_11 * np.power(T11 / t[grid > 500.0], 1 + alpha) * np.exp(-_tau)

    n_grid["H"] = np.concatenate((n_below_500, n_above_500))

    n = {
        s: log_interp1d(grid, n_grid[s])(altitudes)
        for s in ["N2", "O", "O2", "Ar", "He"]
    }

    # Below 150 km, the number density of atomic hydrogen is zero.
    n["H"] = np.concatenate(
        (
            np.zeros(len(altitudes[altitudes < 150.0])),
            log_interp1d(grid[grid >= 150.0], n_grid["H"])(
                altitudes[altitudes >= 150.0]
            ),
        )
    )

    return n


@ureg.wraps(ret=None, args="km", strict=False)
def compute_mean_molar_mass_high_altitude(z):
    r"""
    Computes the mean molar mass in the high-altitude region.

    Parameters
    ----------
    z : array
        Altitude [km].

    Returns
    -------
    ndarray
        Mean molar mass [kg/mol].
    """
    return np.where(z <= 100.0, M0, M["N2"])


@ureg.wraps(ret=None, args="km", strict=False)
def compute_temperature_high_altitude(altitude):
    r"""
    Computes the temperature in the high-altitude region.

    Parameters
    ----------
    altitude : array
        Altitude values [km].

    Returns
    -------
    array:
        Temperature values [K].
    """
    r0 = R0 / 1e3  # km
    a = -76.3232  # K
    b = -19.9429  # km
    tc = 263.1905  # K

    def t(z):
        r"""
        Compute the temperature at a given altitude.

        Parameters
        ----------
        z : float
            Altitude [km].

        Returns
        -------
        float:
            Temperature [K].
        """
        if Z7 <= z <= Z8:
            return T7
        elif Z8 < z <= Z9:
            return tc + a * np.sqrt(1.0 - np.power((z - Z8) / b, 2.0))
        elif Z9 < z <= Z10:
            return T9 + LK9 * (z - Z9)
        elif Z10 < z <= Z12:
            return TINF - (TINF - T10) * np.exp(
                -LAMBDA * (z - Z10) * (r0 + Z10) / (r0 + z)
            )
        else:
            raise ValueError("altitude value is out of range")

    return np.vectorize(t)(altitude)


@ureg.wraps(ret=None, args="km", strict=False)
def compute_temperature_gradient_high_altitude(altitude):
    r"""
    Computes the temperature gradient in the high-altitude region.

    Parameters
    ----------
    altitude : array
        Altitude values [km].

    Returns
    -------
    ndarray
        Temperature gradient values [K/m].
    """
    a = -76.3232  # [dimensionless]
    b = -19.9429  # km

    def gradient(z):
        r"""
        Computes the temperature gradient at a given altitude.

        Parameters
        ----------
        z : float
            Altitude [km].

        Returns
        -------
        float:
            Temperature gradient [K/km].
        """
        if Z7 <= z <= Z8:
            return LK7
        elif Z8 < z <= Z9:
            return -a / b * ((z - Z8) / b) / np.sqrt(1 - np.square((z - Z8) / b))
        elif Z9 < z <= Z10:
            return LK9
        elif Z10 < z <= Z12:
            zeta = (z - Z10) * (R0 + Z10) / (R0 + z)
            return (
                LAMBDA
                * (TINF - T10)
                * np.square((R0 + Z10) / (R0 + z))
                * np.exp(-LAMBDA * zeta)
            )
        else:
            raise ValueError(f"altitude z out of range, should be in " f"[{Z7}, {Z12}]")

    return np.vectorize(gradient)(altitude) / 1e3  # converts K/km to K/m


@ureg.wraps(ret=None, args=("m^-3", "K", "m^-1*s^-1", "m^-1*s^-1"), strict=False)
def thermal_diffusion_coefficient(background, temperature, a, b):
    r"""
    Computes the thermal diffusion coefficient values in the high-altitude
    region.

    Parameters
    ----------
    n : array
        Background number density values [m^-3].

    Parameters
    ----------
    t : array
        Temperature values [K].

    a : float
        Thermal diffusion constant a [m^-1*s^-1].

    b : float
        Thermal diffusion constant b [m^-1*s^-1].

    Returns
    -------
    array
        Values of the thermal diffusion coefficient [m^2/s].
    """
    return (a / background) * np.power(temperature / 273.15, b)


@ureg.wraps(ret=None, args="km", strict=False)
def eddy_diffusion_coefficient(z):
    r"""
    Computes the values of the Eddy diffusion coefficient in the high-altitude
    region.

    Parameters
    ----------
    z : array
        Altitude values [km].

    Returns
    -------
    array
        Eddy diffusion coefficient values [m^2/s].

    Notes
    -----
    Valid in the altitude region :math:`86 <= z <= 150` km.
    """
    return np.where(
        z < 95.0, K_7, K_7 * np.exp(1.0 - (400.0 / (400.0 - np.square(z - 95.0))))
    )


@ureg.wraps(
    ret=None,
    args=("m/s^2", "K", "K/m", "kg/mol", "kg/mol", None, "m^2/s", "m^2/s"),
    strict=False,
)
def f_below_115_km(g, t, dt_dz, m, mi, alpha, d, k):
    r"""
    Evaluates the function :math:`f` defined by equation (36) in
    :cite:`NASA1976USStandardAtmosphere` in the altitude region :math:`86
    <= z <= 115` km.

    Parameters
    ----------
    g : float or array-like
        Values of gravity at the different altitudes [m / s^2].

    t : float or array-like
        Values of temperature at the different altitudes [K].

    dt_dz : float or array-like
        Values of temperature gradient at the different altitudes [K/m].

    m : float
        Molar mass [kg/mol].

    mi : float
        Species molar mass [kg/mol]

    alpha : float
        Alpha thermal diffusion constant [dimensionless].

    d : float or array-like
        Values of the thermal diffusion coefficient at the different altitudes
        [m^2/s].

    k : float or array-like
        Values of the Eddy diffusion coefficient at the different altitudes
        [m^2/s].

    Returns
    -------
    array
        Values of the function f at the different altitudes [m^-1].
    """
    return (g / (R * t)) * (d / (d + k)) * (mi + (m * k) / d + (alpha * R * dt_dz) / g)


@ureg.wraps(ret=None, args=("m/s^2", "K", "K/m", "kg/mol", None), strict=False)
def f_above_115_km(g, t, dt_dz, mi, alpha):
    r"""
    Evaluates the function :math:`f` defined by equation (36) in
    :cite:`NASA1976USStandardAtmosphere` in the altitude region :math:`115 <
    z <= 1000` km.

    Parameters
    ----------
    g : float or array-like
        Values of gravity at the different altitudes [m/s^2].

    t : float or array-like
        Values of temperature at the different altitudes [K].

    dt_dz : float or array-like
        Values of temperature gradient at the different altitudes [K/m].

    mi : float
        Species molar mass [kg/mol]

    alpha : float
        Alpha thermal diffusion constant [dimensionless].

    Returns
    -------
    array
        Values of the function f at the different altitudes [m^-1].
    """
    return (g / (R * t)) * (mi + ((alpha * R) / g) * dt_dz)


@ureg.wraps(
    ret=None,
    args=(None, "km", "m/s^2", "K", "K/m", "kg/mol", "m^2/s", "m^2/s"),
    strict=False,
)
def thermal_diffusion_term(species, grid, g, t, dt_dz, m, d, k):
    r"""
    Computes the thermal diffusion term of a given species in the
    high-altitude region.

    Parameters
    ----------
    species : str
        Species.

    grid : array
        Altitude grid [km].

    g : array
        Values of the gravity on the altitude grid [m/s^2].

    t : array
        Values of the temperature on the altitude grid [K].

    dt_dz : array
        Values of the temperature gradient on the altitude grid [K/m].

    m : array
        Values of the mean molar mass on the altitude grid [kg/mol].

    d : array
        Values of the molecular diffusion coefficient on the altitude grid, for altitudes < 115 km [m^2/s].

    k : array
        Values of the eddy diffusion coefficient on the altitude grid, for altitudes < 115 km [m^2/s].

    Returns
    -------
    array
        Values of the thermal diffusion term [km^-1].
    """
    fo1 = f_below_115_km(
        g[grid < 115.0],
        t[grid < 115.0],
        dt_dz[grid < 115.0],
        m[grid < 115.0],
        M[species],
        ALPHA[species],
        d,
        k,
    )
    fo2 = f_above_115_km(
        g[grid >= 115.0],
        t[grid >= 115.0],
        dt_dz[grid >= 115.0],
        M[species],
        ALPHA[species],
    )
    return np.concatenate((fo1, fo2))


@ureg.wraps(ret=None, args=("km", "m/s^2", "K", "K/m", "m^2/s", "m^2/s"), strict=False)
def thermal_diffusion_term_atomic_oxygen(grid, g, t, dt_dz, d, k):
    r"""
    Computes the thermal diffusion term of atomic oxygen in the high-altitude
    region.

    Parameters
    ----------
    grid : array
        Altitude grid [km].

    g : array
        Values of the gravity on the altitude grid [m/s^2].

    t : array
        Values of the temperature on the altitude grid [K].

    dt_dz : array
        Values of the temperature gradient on the altitude grid [K/m].

    d : array
        Values of thermal diffusion coefficient on the altitude grid [m^2/s].

    k : array
        Values of the Eddy diffusion coefficient on the altitude grid [m^2/s].

    Returns
    -------
    array:
        Values of the thermal diffusion term [km^-1].
    """
    mask1, mask2 = grid < 115.0, grid >= 115.0
    x1 = f_below_115_km(
        g[mask1], t[mask1], dt_dz[mask1], M["N2"], M["O"], ALPHA["O"], d, k
    )
    x2 = f_above_115_km(
        g[grid >= 115.0], t[grid >= 115.0], dt_dz[grid >= 115.0], M["O"], ALPHA["O"]
    )

    return np.concatenate((x1, x2))


@ureg.wraps(
    ret=None, args=("m", "km^-3", "km^-3", "km", "km", "km^-3", "km^-3"), strict=False
)
def velocity_term_hump(z, q1, q2, u1, u2, w1, w2):
    r"""
    Computes the transport term given by equation (37) in
    :cite:`NASA1976USStandardAtmosphere`.

    Parameters
    ----------
    z : float or array-like
        Altitude [km].

    q1 : float
        Value of the Q constant [km^-3].

    q2 : float
        Value of the q constant [km^-3].

    u1 : float
        Value of the U constant [km].

    u2 : floa
        Value of the u constant [km].

    w1 : floa
        Value of the W constant [km^-3].

    w2 : floa
        Value of the w constant [km^-3].

    Returns
    -------
    float or array-like:
        Values of the transport term [km^-1].

    Notes
    -----
    Valid in the altitude region: 86 km <= z <= 150 km
    """
    # @formatter:off
    return (
        q1 * np.square(z - u1) * np.exp(-w1 * np.power(z - u1, 3.0))
        + q2 * np.square(u2 - z) * np.exp(-w2 * np.power(u2 - z, 3.0))
    ) / 1e3  # the factor 1e3 converts m^-1 to km^-1
    # @formatter:on


@ureg.wraps(ret=None, args=("km", "km^-3", "km^-3", "km^-3"), strict=False)
def velocity_term_no_hump(z, q1, u1, w1):
    r"""
    Computes the transport term given by equation (37) in
    :cite:`NASA1976USStandardAtmosphere` where the second term is zero.

    Parameters
    ----------
    z : float or array-like
        Altitude [km].

    q1 : float
        Value of the Q constant [km^-3].

    u1 : float
        Value of the U constant [km].

    w1 : floa
        Value of the W constant [km^-3].

    Returns
    -------
    float or array-like:
        Values of the transport term [km^-1].

    Notes
    -----
    Valid in the altitude region :math:`86 <= z <= 150` km.
    """
    # @formatter:off
    return (
        q1 * np.square(z - u1) * np.exp(-w1 * np.power(z - u1, 3.0))
    ) / 1e3  # the factor 1e3 converts m^-1 to km^-1
    # @formatter:on


@ureg.wraps(ret=None, args=(None, "km"), strict=False)
def velocity_term(species, grid):
    r"""
    Computes the velocity term of a given species in the high-altitude region.

    Parameters
    ----------
    species : str
        Species.

    grid : array
        Altitude grid [km].

    Returns
    -------
    array:
        Values of the velocity terms [km^-1].

    Notes
    -----
    Not valid for atomic oxygen. See :func:`velocity_term_atomic_oxygen`
    """
    x1 = velocity_term_no_hump(
        grid[grid <= 150.0], Q1[species], U1[species], W1[species]
    )

    # Above 150 km, the velocity term is neglected, as indicated at p. 14 in
    # :cite:`NASA1976USStandardAtmosphere`
    x2 = np.zeros(len(grid[grid > 150.0]))

    return np.concatenate((x1, x2))


@ureg.wraps(ret=None, args="km", strict=False)
def velocity_term_atomic_oxygen(grid):
    r"""
    Computes the velocity term of atomic oxygen in the high-altitude region.

    Parameters
    ----------
    grid : array
        Altitude grid [km].

    Returns
    -------
    array:
        Values of the velocity term [km^-1].
    """
    mask1, mask2 = grid <= 150.0, grid > 150.0
    x1 = np.where(
        grid[mask1] <= 97.0,
        velocity_term_hump(
            grid[mask1], Q1["O"], Q2["O"], U1["O"], U2["O"], W1["O"], W2["O"]
        ),
        velocity_term_no_hump(grid[mask1], Q1["O"], U1["O"], W1["O"]),
    )

    x2 = np.zeros(len(grid[mask2]))
    return np.concatenate((x1, x2))


@ureg.wraps(ret=None, args=("km", None), strict=False)
def tau_function(z_grid, below_500=True):
    r"""
    Computes the integral given by equation (40) in
    :cite:`NASA1976USStandardAtmosphere` at each point of an altitude grid.

    Parameters
    ----------
    z_grid : array-like
        Altitude grid (values sorted by ascending order) to use for integration [km].

    below_500 : bool
        True if altitudes in z_grid are lower than 500 km, False otherwise.

    Returns
    -------
    array:
        Integral evaluations [dimensionless].

    Notes
    -----
    Valid for altitudes between 150 km and 500 km.
    """
    if below_500:
        z_grid = z_grid[::-1]

    y = (
        M["H"]
        * compute_gravity(ureg.Quantity(z_grid, "km"))
        / (R * compute_temperature_high_altitude(z_grid))
    )  # [m^-1]
    integral_values = cumtrapz(y, 1e3 * z_grid, initial=0.0)  # the factor 1e3
    # converts z_grid to meters

    if below_500:
        return integral_values[::-1]
    else:
        return integral_values


def log_interp1d(x, y):
    """
    Computes the linear interpolation of :math:`y(x)` in logarithmic space.

    Parameters
    ----------
    x : array
        1-D array of real value.

    y : array
        N-D array of real values. The length of y along the interpolation axis
        must be equal to the length of x.

    Returns
    -------
    callable:
        Function whose call method uses interpolation to find the value of new
        points.
    """
    logx = np.log10(x)
    logy = np.log10(y)
    lin_interp = interp1d(logx, logy, kind="linear")

    def log_interp(z):
        return np.power(10.0, lin_interp(np.log10(z)))

    return log_interp


@ureg.wraps(ret=None, args=("m", "Pa", "K"), strict=False)
def compute_pressure_low_altitude(h, pb, tb):
    r"""
    Computes the pressure in the low-altitude region.

    Parameters
    ----------
    h : array
        Geopotential height values [m].

    p_levels : array-like
        Levels pressure [Pa].

    Returns
    -------
    array:
        Pressure values [Pa].
    """
    # we create a mask for each layer
    masks = [ma.masked_inside(h, H[i - 1], H[i]).mask for i in range(1, 8)]

    # for each layer, we evaluate the pressure based on whether the
    # temperature gradient is zero or not
    p = np.empty(len(h))
    for i, mask in enumerate(masks):
        if LK[i] == 0:
            p[mask] = compute_pressure_low_altitude_zero_gradient(
                h[mask], H[i], pb[i], tb[i]
            )
        else:
            p[mask] = compute_pressure_low_altitude_non_zero_gradient(
                h[mask], H[i], pb[i], tb[i], LK[i]
            )
    return p


@ureg.wraps(ret=None, args=("m", "m", "Pa", "K"), strict=False)
def compute_pressure_low_altitude_zero_gradient(h, hb, pb, tb):
    r"""
    Computes the pressure in the low-altitude region when the temperature
    gradient is zero.

    Parameters
    ----------
    h : float or array-like
        Geopotential height [m].

    hb : float or array-like
        Geopotential height at the bottom of the layer [m].

    pb : float or array-like
        Pressure at the bottom of the layer [Pa].

    tb : float or array-like
        Temperature at the bottom of the layer [K].

    Returns
    -------
    float or array-like:
        Pressure [Pa].
    """
    return pb * np.exp(-G0 * M0 * (h - hb) / (R * tb))


@ureg.wraps(ret=None, args=("m", "m", "Pa", "K", "K/m"), strict=False)
def compute_pressure_low_altitude_non_zero_gradient(h, hb, pb, tb, lkb):
    r"""
    Computes the pressure in the low-altitude region when the temperature
    gradient is non-zero.

    Parameters
    ----------
    h : float or array-like
        Geopotential height [m].

    hb : float or array-like
        Geopotential height at the bottom of the layer [m].

    pb : float or array-like
        Pressure at the bottom of the layer [Pa].

    tb : float or array-like
        Temperature at the bottom of the layer [K].

    Returns
    -------
    float or array-like:
        Pressure [Pa].
    """
    return pb * np.power(tb / (tb + lkb * (h - hb)), G0 * M0 / (R * lkb))


@ureg.wraps(ret=None, args=("m", "K"), strict=False)
def compute_temperature_low_altitude(h, tb):
    r"""
    Computes the temperature in the low-altitude region.

    Parameters
    ----------
    h : array
        Geopotential height values [m].

    tb : array-like
        Levels temperature values [K].

    Returns
    -------
    array:
        Temperature [K].
    """
    # we create a mask for each layer
    masks = [ma.masked_inside(h, H[i - 1], H[i]).mask for i in range(1, 8)]

    # for each layer, we evaluate the pressure based on whether the
    # temperature gradient is zero or not
    t = np.empty(len(h))
    for i, mask in enumerate(masks):
        if LK[i] == 0:
            t[mask] = tb[i]
        else:
            t[mask] = tb[i] + LK[i] * (h[mask] - H[i])
    return t


@ureg.wraps(ret=None, args="m", strict=False)
def to_altitude(h):
    r"""
    Converts geopotential height to (geometric) altitude.

    Parameters
    ----------
    h : float or array-like
        Geopotential altitude [m].

    Returns
    -------
    float or array-like:
        Altitude [m]
    """
    return R0 * h / (R0 - h)


@ureg.wraps(ret=None, args="m", strict=False)
def to_geopotential_height(z):
    r"""
    Converts altitude to geopotential height.

    Parameters
    ----------
    z : float or array-like
        Altitude [m].

    Returns
    -------
    float or array-like:
        Geopotential height [m]
    """
    return R0 * z / (R0 + z)


@ureg.wraps(ret=None, args="m", strict=False)
def compute_gravity(z):
    r"""
    Computes the gravity.

    Parameters
    ----------
    z : float or array-like
        Altitude [m].

    Returns
    -------
    float or array-like:
        Gravity [m/s^2].
    """
    return G0 * np.power((R0 / (R0 + z)), 2.0)



from __future__ import annotations

import logging

import attrs
import numpy as np
import xarray as xr
from pinttr.util import always_iterable

import eradiate

from ._core import PipelineStep
from ..attrs import documented, parse_docs
from ..exceptions import UnsupportedModeError
from ..kernel import bitmap_to_dataset
from ..units import symbol
from ..units import unit_context_config as ucc

logger = logging.getLogger(__name__)


def _spectral_dims():
    if eradiate.mode().is_mono:
        return (
            (
                "w",
                {
                    "standard_name": "radiation_wavelength",
                    "long_name": "wavelength",
                    "units": symbol(ucc.get("wavelength")),
                },
            ),
        )
    elif eradiate.mode().is_ckd:
        return (
            ("bin", {"standard_name": "ckd_bin", "long_name": "CKD bin"}),
            ("index", {"standard_name": "ckd_index", "long_name": "CKD index"}),
        )
    else:
        raise UnsupportedModeError


@parse_docs
@attrs.define
class Gather(PipelineStep):
    """
    Gather raw kernel results (output as nested dictionaries) into an xarray
    dataset.

    This pipeline step takes a nested dictionary produced by the parametric loop
    of an :class:`.Experiment` and repackages it as a :class:`~xarray.Dataset`.
    The top-level spectral index is mapped to mode-dependent spectral
    coordinates. Film dimensions are left unmodified and retain their
    metadata.

    An ``img`` variable holds sensor values. An ``spp`` variable holds the
    sample count.
    """

    var: str | tuple[str, dict] = documented(
        attrs.field(default="img"),
        default='"img"',
        type="str or tuple[str, dict]",
        init_type="str or tuple[str, dict], optional",
        doc="Name of the variable containing sensor data. Optionally, a "
        "(name, metadata) pair can be passed.",
    )

    def transform(self, x: dict) -> xr.Dataset:
        logger.debug("gather: begin")
        # Basic preparation
        spectral_dims = []
        spectral_dim_metadata = {}

        for y in _spectral_dims():
            if isinstance(y, str):
                spectral_dims.append(y)
                spectral_dim_metadata[y] = {}
            else:
                spectral_dims.append(y[0])
                spectral_dim_metadata[y[0]] = y[1]

        sensor_datasets = []

        # Loop on spectral indexes
        for siah, result_dict in x.items():
            if eradiate.mode().is_mono:
                spectral_index = siah
            elif eradiate.mode().is_ckd:
                spectral_index = (
                    str(int(siah[0])),  # TODO: PR#311 hack
                    siah[1],
                )

            ds = bitmap_to_dataset(result_dict["bitmap"])
            spp = result_dict["spp"]

            # Set spectral coordinates
            all_coords = {
                spectral_dim: [spectral_coord]
                for spectral_dim, spectral_coord in zip(
                    spectral_dims, always_iterable(spectral_index)
                )
            }

            # Add spectral and sensor dimensions to img array
            ds["img"] = ds.img.expand_dims(dim=all_coords)

            # Package spp in a data array
            all_dims = list(all_coords.keys())
            ds["spp"] = (all_dims, np.reshape(spp, [1 for _ in all_dims]))

            sensor_datasets.append(ds)

        # Combine all the data
        with xr.set_options(keep_attrs=True):
            result = xr.combine_by_coords(sensor_datasets)

        # Drop "channel" dimension when using a monochromatic Mitsuba variant
        if eradiate.mode().check(mi_color_mode="mono"):
            result = result.squeeze("channel", drop=True)

        for spectral_dim in spectral_dims:
            result[spectral_dim].attrs = spectral_dim_metadata[spectral_dim]

        # Apply metadata to data variables
        if isinstance(self.var, str):
            var = self.var
            var_metadata = {}
        else:
            var = self.var[0]
            var_metadata = self.var[1]

        result = result.rename({"img": var})
        result[var].attrs.update(var_metadata)

        logger.debug("gather: end")

        return result



from __future__ import annotations

import logging
import typing as t

import attrs
import numpy as np
from pinttr.util import always_iterable

from ._core import PipelineStep
from ..attrs import documented, parse_docs
from ..scenes.measure import Measure
from ..scenes.spectra import InterpolatedSpectrum, UniformSpectrum
from ..units import symbol, to_quantity
from ..units import unit_registry as ureg

logger = logging.getLogger(__name__)


@parse_docs
@attrs.define
class ApplySpectralResponseFunction(PipelineStep):
    r"""
    Apply spectral response function to specified variables.

    This post-processing pipeline step applies the spectral response function
    to specified variables. It creates new corresponding data variables with no
    dependency against the wavelength dimension.

    Notes
    -----
    The processed dataset is expected to have a ``bin`` coordinate, associated
    with bounds ``bin_wmin`` and ``bin_wmax``. If not, :meth:`transform` will
    raise an exception.

    A common nomenclature practice refers to this process as "convolution"
    :cite:`Burggraaff2020BiasesIncorrectReflectance`,
    but we prefer to avoid this term as we consider it to refer to
    another mathematical operation defined by:

    .. math::

        (f \ast g) \, (t) := \int_{-\infty}^{+\infty} f(\tau) \, g(t - \tau) \, d\tau

    Instead, we adopt the term "spectral response weighted average" since the
    operation is defined by:

    .. math::

        \overline{x} = \frac{\int x(\lambda) \, w(\lambda) \, d\lambda}
                    {\int w(\lambda) \, d\lambda}

    where

    * :math:`x` is the variable to be weighted,
    * :math:`w` is the spectral response function,

    and which effectively translates into a weighted average in numerical form,
    with the weights being the spectral response function values.
    """

    measure: Measure = documented(
        attrs.field(
            validator=attrs.validators.instance_of(Measure),
            repr=lambda self: f"{self.__class__.__name__}(id='{self.id}', ...)",
        ),
        doc="A :class:`.Measure` instance from which the processed data originates.",
        type=":class:`.Measure`",
    )

    vars: list[str] = documented(
        attrs.field(
            factory=list,
            converter=lambda x: list(always_iterable(x)),
            validator=attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(str)
            ),
        ),
        doc="List of variables to which the spectral response function is to be "
        "applied.",
        type="list of str",
        init_type="str or list of str",
        default="[]",
    )

    def transform(self, x: t.Any) -> t.Any:
        result = x.copy(deep=False)

        if not {"bin_wmin", "bin_wmax"}.issubset(set(result.coords.keys())):
            raise ValueError(
                "input data is missing 'bin_wmin' and/or 'bin_wmax' coordinates"
            )

        measure = self.measure

        # Evaluate integral of spectral response function within selected interval
        wmin = to_quantity(result.bin_wmin).min()
        wmax = to_quantity(result.bin_wmax).max()
        srf = measure.srf
        srf_int = srf.integral(wmin, wmax)

        if isinstance(srf, InterpolatedSpectrum):
            srf_w = srf.wavelengths
        elif isinstance(srf, UniformSpectrum):
            srf_w = np.array([wmin.m_as(ureg.nm), wmax.m_as(ureg.nm)]) * ureg.nm
        else:
            raise TypeError(f"unhandled SRF type '{srf.__class__.__name__}'")

        for var in self.vars:
            # Evaluate integral of product of variable and SRF within selected interval
            data_w = to_quantity(result.w)

            # Spectral grid is the finest between data and SRF grids
            w_units = data_w.units
            w_m = np.array(sorted(set(data_w.m_as(w_units)) | set(srf_w.m_as(w_units))))

            # If data var has length 1 on spectral dimension, directly select
            # the value instead of using interpolation (it's a known scipy issue)
            if len(result[var].w) == 1:
                # Note: The tricky thing is to recreate and extend the 'w'
                # dimension with the same axis index as in the original data
                var_values = (
                    result[var]
                    .isel(w=0, drop=True)
                    .expand_dims(w=w_m, axis=result[var].get_axis_num("w"))
                )

            # Otherwise, use nearest neighbour interpolation (we assume that var
            # is constant over each spectral bin)
            else:
                var_values = result[var].interp(
                    w=w_m, method="nearest", kwargs={"fill_value": "extrapolate"}
                )

            srf_values = (
                srf.eval_mono(w_m * w_units)
                .reshape([-1 if dim == "w" else 1 for dim in var_values.dims])
                .magnitude
            )
            assert isinstance(srf_values, np.ndarray)  # Check for leftover bugs
            var_srf_int = (var_values * srf_values).integrate("w")

            # Apply SRF to variable and store result
            dims = list(result[var].dims)
            dims.remove("w")
            result[f"{var}_srf"] = (dims, var_srf_int.values / srf_int.m_as(w_units))
            attrs = result[var].attrs.copy()

            if "standard_name" in attrs:
                attrs["standard_name"] += "_srf"
            if "long_name" in attrs:
                attrs["long_name"] += " (SRF applied)"
            result[f"{var}_srf"].attrs = attrs

        logger.debug("ApplySpectralResponseFunction pipeline step: end")

        return result


@parse_docs
@attrs.define
class ComputeReflectance(PipelineStep):
    """
    Derive reflectance from radiance and irradiance values.
    """

    radiance_var: str = documented(
        attrs.field(default="radiance", validator=attrs.validators.instance_of(str)),
        doc="Name of the variable storing leaving radiance values.",
        type="str",
        default='"radiance"',
    )

    irradiance_var: str = documented(
        attrs.field(default="irradiance", validator=attrs.validators.instance_of(str)),
        doc="Name of the variable storing irradiance (incoming flux) values.",
        type="str",
        default='"irradiance"',
    )

    brdf_var: str = documented(
        attrs.field(default="brdf", validator=attrs.validators.instance_of(str)),
        doc="Name of the variable storing BRDF values.",
        type="str",
        default='"brdf"',
    )

    brf_var: str = documented(
        attrs.field(default="brf", validator=attrs.validators.instance_of(str)),
        doc="Name of the variable storing BRF values.",
        type="str",
        default='"brf"',
    )

    def transform(self, x: t.Any) -> t.Any:
        logger.debug("ComputeReflectance pipeline step: begin")
        # Compute BRDF and BRF
        result = x.copy(deep=False)

        # We assume that all quantities are stored in kernel units
        result[self.brdf_var] = result[self.radiance_var] / result[self.irradiance_var]
        result[self.brdf_var].attrs = {
            "standard_name": "brdf",
            "long_name": "bi-directional reflection distribution function",
            "units": symbol("1/sr"),
        }

        result[self.brf_var] = result[self.brdf_var] * np.pi
        result[self.brf_var].attrs = {
            "standard_name": "brf",
            "long_name": "bi-directional reflectance factor",
            "units": symbol("dimensionless"),
        }
        logger.debug("ComputeReflectance pipeline step: end")

        return result


@parse_docs
@attrs.define
class ComputeAlbedo(PipelineStep):
    """
    Derive the albedo from radiosity and irradiance fields.
    """

    radiosity_var: str = documented(
        attrs.field(default="radiosity", validator=attrs.validators.instance_of(str)),
        doc="Name of the variable storing the radiosity (leaving flux) value.",
        type="str",
        default='"radiosity"',
    )

    irradiance_var: str = documented(
        attrs.field(default="irradiance", validator=attrs.validators.instance_of(str)),
        doc="Name of the variable storing the irradiance (incoming flux) value.",
        type="str",
        default='"irradiance"',
    )

    albedo_var: str = documented(
        attrs.field(default="albedo", validator=attrs.validators.instance_of(str)),
        doc="Name of the variable storing the albedo value.",
        type="str",
        default='"albedo"',
    )

    def transform(self, x: t.Any) -> t.Any:
        logger.debug("ComputeAlbedo pipeline step: begin")
        # Compute albedo
        result = x.copy(deep=False)

        # We assume that all quantities are stored in kernel units
        result[self.albedo_var] = (
            result[self.radiosity_var] / result[self.irradiance_var]
        )
        result[self.albedo_var].attrs = {
            "standard_name": "albedo",
            "long_name": "surface albedo",
            "units": "",
        }
        logger.debug("ComputeAlbedo pipeline step: end")

        return result



import numpy as np
import pytest
import xarray as xr

from eradiate import unit_registry as ureg
from eradiate.radprops.absorption import compute_sigma_a
from eradiate.units import to_quantity


def test_compute_sigma_a_no_t_coord():
    # data set with no t coordinate
    ds = xr.Dataset(
        data_vars={"xs": (("w", "p"), np.random.random((4, 3)), dict(units="cm^2"))},
        coords={
            "w": ("w", np.linspace(18000, 18010, 4), dict(units="cm^-1")),
            "p": ("p", np.linspace(90000, 110000, 3), dict(units="Pa")),
        },
    )

    # returns a single absorption coefficient value when input pressure is
    # a scalar
    wl = ureg.Quantity(555.5, "nm")
    x = compute_sigma_a(ds, wl=wl)
    assert isinstance(x, ureg.Quantity)
    assert x.check("[length]^-1")
    assert len(x) == 1

    # handles multiple wavelength values
    x = compute_sigma_a(ds, wl=ureg.Quantity(np.array([555.50, 555.51]), "nm"))
    assert len(x) == 2

    # does not raise when 'fill_value' is used
    x = compute_sigma_a(
        ds, wl=wl, p=ureg.Quantity(120000.0, "Pa"), fill_values=dict(pt=0.0)
    )

    # raises when wavelength out of range
    with pytest.raises(ValueError):
        compute_sigma_a(ds, wl=ureg.Quantity(560.0, "nm"))

    # raises when pressure out of range
    with pytest.raises(ValueError):
        compute_sigma_a(ds, wl=wl, p=ureg.Quantity(120000.0, "Pa"))

    # returns an array of absorption coefficient values when input pressure is
    # an array
    ds_p = to_quantity(ds.p)
    x = compute_sigma_a(ds=ds, wl=wl, p=np.linspace(ds_p.min(), ds_p.max(), num=10))
    assert isinstance(x, ureg.Quantity)
    assert x.check("[length]^-1")
    assert len(x) == 10

    # absorption coefficient scales with number density
    x = compute_sigma_a(ds, wl=wl, n=ureg.Quantity(np.array([1.0, 2.0]), "m^-3"))
    assert x[1] == 2 * x[0]

    # dataset with no temperature coordinate is not interpolated on temperature
    x = compute_sigma_a(
        ds, wl=wl, t=ureg.Quantity(200.0, "K"), n=ureg.Quantity(1.0, "m^-3")
    )
    y = compute_sigma_a(
        ds, wl=wl, t=ureg.Quantity(300.0, "K"), n=ureg.Quantity(1.0, "m^-3")
    )
    assert x == y


def test_compute_sigma_a_t_coord():
    # data set with t coordinate
    ds = xr.Dataset(
        data_vars={
            "xs": (("w", "p", "t"), np.random.random((4, 3, 2)), dict(units="cm^2"))
        },
        coords={
            "w": ("w", np.linspace(18000, 18010, 4), dict(units="cm^-1")),
            "p": ("p", np.linspace(90000, 110000, 3), dict(units="Pa")),
            "t": ("t", np.array([250, 300]), dict(units="K")),
        },
    )

    # returns a single absorption coefficient value when input pressure
    # and temperature are scalars
    wl = ureg.Quantity(555.5, "nm")
    x = compute_sigma_a(ds, wl=wl)
    assert isinstance(x, ureg.Quantity)
    assert x.check("[length]^-1")
    assert len(x) == 1

    # handles multiple wavelength values
    x = compute_sigma_a(ds, wl=ureg.Quantity(np.array([555.50, 555.51]), "nm"))
    assert len(x) == 2

    # does not raise when 'fill_value' is used
    x = compute_sigma_a(
        ds, wl=wl, p=ureg.Quantity(120000.0, "Pa"), fill_values=dict(pt=0.0)
    )

    # raises when wavelength out of range
    with pytest.raises(ValueError):
        compute_sigma_a(ds, wl=ureg.Quantity(560.0, "nm"))

    # raises when pressure out of range
    with pytest.raises(ValueError):
        compute_sigma_a(ds, wl=wl, p=ureg.Quantity(120000.0, "Pa"))

    # returns an array of absorption coefficient values when input pressure is
    # an array
    ds_p = to_quantity(ds.p)
    x = compute_sigma_a(ds=ds, wl=wl, p=np.linspace(ds_p.min(), ds_p.max(), num=10))
    assert isinstance(x, ureg.Quantity)
    assert x.check("[length]^-1")
    assert len(x) == 10
    assert not any(np.isnan(x.magnitude))

    # returns a 1D array of absorption coefficient values when input pressure
    # and temperature are arrays
    ds_t = to_quantity(ds.t)
    x = compute_sigma_a(
        ds=ds,
        wl=wl,
        p=np.linspace(ds_p.min(), ds_p.max(), num=12),
        t=np.linspace(ds_t.min(), ds_t.max(), num=12),
    )
    assert isinstance(x, ureg.Quantity)
    assert x.check("[length]^-1")
    assert len(x) == 12
    assert not any(np.isnan(x.magnitude))

    # raises when input pressure and temperature are arrays with different
    # lengths
    with pytest.raises(ValueError):
        x = compute_sigma_a(
            ds=ds,
            wl=wl,
            p=np.linspace(ds_p.min(), ds_p.max(), num=12),
            t=np.linspace(ds_t.min(), ds_t.max(), num=11),
        )

    # absorption coefficient scales with number density
    x = compute_sigma_a(ds, wl=wl, n=ureg.Quantity(np.array([1.0, 2.0]), "m^-3"))
    assert x[1] == 2 * x[0]

    # dataset with temperature coordinate is interpolated on temperature
    x = compute_sigma_a(
        ds, wl=wl, t=ureg.Quantity(260.0, "K"), n=ureg.Quantity(1.0, "m^-3")
    )
    y = compute_sigma_a(
        ds, wl=wl, t=ureg.Quantity(280.0, "K"), n=ureg.Quantity(1.0, "m^-3")
    )
    assert x != y



"""
Functions to compute monochromatic absorption.
"""
from __future__ import annotations

import numpy as np
import pint
import xarray as xr
from scipy.constants import physical_constants

from ..units import to_quantity
from ..units import unit_registry as ureg

_BOLTZMANN = ureg.Quantity(*physical_constants["Boltzmann constant"][:2])


def compute_sigma_a(
    ds: xr.Dataset,
    wl: pint.Quantity = ureg.Quantity(550.0, "nm"),
    p: pint.Quantity = ureg.Quantity(101325.0, "Pa"),
    t: pint.Quantity = ureg.Quantity(288.15, "K"),
    n: pint.Quantity | None = None,
    fill_values: float | None = None,
    methods: dict[str, str] | None = None,
) -> pint.Quantity:
    """
    Compute monochromatic absorption coefficient at given wavelength,
    pressure and temperature values.

    Parameters
    ----------
    ds : Dataset
        Absorption cross-section data set.

    wl : quantity
        Wavelength [nm].

    p : quantity
        Pressure [Pa].

        .. note:: If ``p``, ``t`` and ``n`` are arrays, their lengths must be
           the same.

    t : quantity
        Temperature [K].

        .. note:: If the coordinate ``t`` is not in the input dataset ``ds``,
           the interpolation on temperature is not performed.

    n : quantity
        Number density [m^-3].

        .. note:: If ``n`` is ``None``, the values of ``t`` and ``p`` are then
           used only to compute the corresponding number density.

    fill_values : dict, optional
        Mapping of coordinates (in ``["w", "pt"]``) and fill values (either
        ``None`` or float).
        If not ``None``, out of bounds values are assigned the fill value
        during interpolation along the wavelength or pressure and temperature
        coordinates.
        If ``None``, out of bounds values will trigger the raise of a
        ``ValueError``.
        Only one fill value can be provided for both pressure and temperature
        coordinates.

    methods : dict, optional
        Mapping of coordinates (in ``["w", "pt"]``) and interpolation methods.
        Default interpolation method is linear.
        Only one interpolation method can be specified for both pressure
        and temperature coordinates.

    Returns
    -------
    quantity
        Absorption coefficient values.

    Raises
    ------
    ValueError
        When wavelength, pressure, or temperature values are out of the range
        of the data set and the corresponding fill value in ``fill_values`` is
        ``None``.

    Warnings
    --------
    The values of the absorption cross-section at the desired wavelength,
    pressure and temperature values,
    :math:`\\sigma_{a\\lambda} (p, T)`,
    are obtained by interpolating the input absorption cross-section data
    set along the corresponding dimensions.

    Notes
    -----
    The absorption coefficient is given by:

    .. math::
        k_{a\\lambda} = n \\, \\sigma_{a\\lambda} (p, T)

    where

    * :math:`k_{a\\lambda}` is the absorption coefficient [:math:`L^{-1}`],
    * :math:`\\lambda` is the wavelength [:math:`L`],
    * :math:`n` is the number density [:math:`L^{-3}`],
    * :math:`\\sigma_a` is the absorption cross section [:math:`L^2`],
    * :math:`p` is the pressure [:math:`ML^{-1}T^{-2}`] and
    * :math:`t` is the temperature [:math:`\\Theta`].
    """

    if fill_values is None:
        fill_values = dict(w=None, pt=None)

    if methods is None:
        methods = dict(w="linear", pt="linear")

    for name in ["w", "pt"]:
        if name not in fill_values:
            fill_values[name] = None
        if name not in methods:
            methods[name] = "linear"

    # Interpolate along wavenumber dimension
    xsw = ds.interp(
        w=(1.0 / wl).m_as(ds.w.units),  # wavenumber in cm^-1
        method=methods["w"],
        kwargs=dict(
            bounds_error=(fill_values["w"] is None),
            fill_value=fill_values["w"],
        ),
    )

    # If the data set includes a temperature coordinate, we interpolate along
    # both pressure and temperature dimensions.
    # Else, we interpolate only along the pressure dimension.
    p_m = p.m_as(ds.p.units)
    p_values = np.array([p_m]) if isinstance(p_m, float) else p_m
    pz = xr.DataArray(p_values, dims="pt")
    if "t" in ds.coords:
        t_m = t.m_as(ds.t.units)
        t_values = np.array([t_m] * len(p_values)) if isinstance(t_m, float) else t_m
        tz = xr.DataArray(t_values, dims="pt")
        interpolated = xsw.interp(
            p=pz,
            t=tz,
            method=methods["pt"],
            kwargs=dict(
                bounds_error=(fill_values["pt"] is None),
                fill_value=fill_values["pt"],
            ),
        )
    else:
        interpolated = xsw.interp(
            p=pz,
            method=methods["pt"],
            kwargs=dict(
                bounds_error=(fill_values["pt"] is None),
                fill_value=fill_values["pt"],
            ),
        )

    xs = to_quantity(interpolated.xs)

    n = p / (_BOLTZMANN * t) if n is None else n

    return (n * xs).to("km^-1")



"""
Atmospheric thermophysical properties profiles models according to
:cite:`Anderson1986AtmosphericConstituentProfiles`.
"""
from __future__ import annotations

import typing as t

import pint
import xarray as xr

from .util import compute_scaling_factors, interpolate, rescale_concentration
from .. import data
from .._mode import mode

VALID_MODEL_IDS = [
    "tropical",
    "midlatitude_summer",
    "midlatitude_winter",
    "subarctic_summer",
    "subarctic_winter",
    "us_standard",
]


def make_profile(
    model_id: str = "us_standard",
    levels: pint.Quantity | None = None,
    concentrations: t.MutableMapping[str, pint.Quantity] | None = None,
) -> xr.Dataset:
    """
    Return atmosphere thermophysical properties based on atmosphere models
    defined in AFGL's 1986 technical report
    :cite:`Anderson1986AtmosphericConstituentProfiles`.

    This function lets you interpolate the atmosphere thermophysical properties
    on a different level altitude mesh than the original.

    You can also rescale the molecules' fractions so that target concentrations
    are met for these molecules.
    Target concentrations can be provided in different units.
    For more information about rescaling process and the supported
    concentration units, refer to the documentation of
    :func:`~eradiate.thermoprops.util.compute_scaling_factors`.

    Parameters
    ----------
    model_id : {"us_standard", "midlatitude_summer", "midlatitude_winter", \
        "subarctic_summer", "subarctic_winter", "tropical"}, \
        optional, default: "us_standard"
        Model identifier.

    levels : quantity or array, optional
        Altitude levels. The array must contain at least two values.
        If not provided, the atmospheric profile is built using the data set's
        altitude levels.

    concentrations : dict, optional
        Molecules concentrations as a {str: quantity} mapping.

    Returns
    -------
    Dataset
        Atmosphere thermophysical properties.

    Notes
    -----
    :cite:`Anderson1986AtmosphericConstituentProfiles` defines six models,
    listed in the table below.

    .. list-table:: AFGL (1986) atmospheric thermophysical properties models
       :widths: 2 4 4
       :header-rows: 1

       * - Model number
         - Model identifier
         - Model name
       * - 1
         - ``tropical``
         - Tropic (15N Annual Average)
       * - 2
         - ``midlatitude_summer``
         - Mid-Latitude Summer (45N July)
       * - 3
         - ``midlatitude_winter``
         - Mid-Latitude Winter (45N Jan)
       * - 4
         - ``subarctic_summer``
         - Sub-Arctic Summer (60N July)
       * - 5
         - ``subarctic_winter``
         - Sub-Arctic Winter (60N Jan)
       * - 6
         - ``us_standard``
         - U.S. Standard (1976)

    .. attention::
       The original altitude mesh specified by
       :cite:`Anderson1986AtmosphericConstituentProfiles` is a piece-wise
       regular altitude mesh with an altitude step of 1 km from 0 to 25 km,
       2.5 km from 25 km to 50 km and 5 km from 50 km to 120 km.
       Since the Eradiate kernel only supports regular altitude mesh, the
       original atmospheric thermophysical properties profiles were
       interpolated on the regular altitude mesh with an altitude step of 1 km
       from 0 to 120 km.

    All six models include the following 10 molecular species:
    H2O, CO2, O3, N2O, CO, CH4, O2, NO, SO2 and NO2.
    """
    if model_id not in VALID_MODEL_IDS:
        raise ValueError(f"model_id should be in {VALID_MODEL_IDS} (got '{model_id}')")
    species = set(concentrations.keys()) if concentrations else set()
    unhandled = species - {"H2O", "CO2", "O3"}

    if unhandled:
        if mode().is_ckd:
            raise NotImplementedError(
                f"molecules '{unhandled}' cannot be rescaled in ckd mode"
            )

    thermoprops = data.load_dataset(f"thermoprops/afgl_1986-{model_id}.nc")

    if levels is not None:
        thermoprops = interpolate(
            ds=thermoprops,
            z_level=levels,
            conserve_columns=True,
        )

    if concentrations is not None:
        factors = compute_scaling_factors(
            ds=thermoprops,
            concentration=concentrations,
        )
        thermoprops = rescale_concentration(ds=thermoprops, factors=factors)

    return thermoprops



"""Test cases of the _molecular_atmosphere module."""

import mitsuba as mi
import numpy.testing as npt
import pytest

from eradiate import KernelContext
from eradiate import unit_registry as ureg
from eradiate.scenes.atmosphere import MolecularAtmosphere
from eradiate.scenes.core import Scene, traverse
from eradiate.test_tools.types import check_scene_element
from eradiate.test_tools.util import skipif_data_not_found


@pytest.fixture
def absorption_dataset_550nm():
    return "spectra/absorption/us76_u86_4/us76_u86_4-spectra-18000_19000.nc"


def test_molecular_atmosphere_default(mode_mono):
    """Default  constructor produces a valid kernel dictionary."""
    skipif_data_not_found(
        f"spectra/absorption/us76_u86_4/us76_u86_4-spectra-18000_19000.nc"
    )
    atmosphere = MolecularAtmosphere()
    check_scene_element(atmosphere)


def test_molecular_atmosphere_scale(mode_mono):
    skipif_data_not_found(
        f"spectra/absorption/us76_u86_4/us76_u86_4-spectra-18000_19000.nc"
    )
    atmosphere = MolecularAtmosphere(scale=2.0)
    template, _ = traverse(atmosphere)
    kernel_dict = template.render(KernelContext())
    assert kernel_dict["medium_atmosphere"]["scale"] == 2.0


def test_molecular_atmosphere_afgl_1986(mode_ckd):
    # afgl_1986() constructor produces a valid kernel dictionary
    atmosphere = MolecularAtmosphere.afgl_1986()
    template, params = traverse(Scene(objects={"atmosphere": atmosphere}))
    mi_scene: mi.Scene = mi.load_dict(template.render(KernelContext()))

    # CKD evaluation generates valid parameter update tables
    mi_params: mi.SceneParameters = mi.traverse(mi_scene)

    for w in [280, 550, 1040, 2120, 2400] * ureg.nm:
        ctx = KernelContext(si={"w": w, "g": 0.3})
        mi_params.update(params.render(ctx))


@pytest.mark.slow
def test_molecular_atmosphere_ussa_1976(mode_mono):
    # ussa_1976() constructor produces a valid kernel dictionary

    for wavenumber_range in [
        "15000_16000",
        "18000_19000",
        "22000_23000",
    ]:
        skipif_data_not_found(
            f"spectra/absorption/us76_u86_4/us76_u86_4-spectra-{wavenumber_range}.nc"
        )

    eval_w = [440.0, 550.0, 660.0] * ureg.nm
    atmosphere = MolecularAtmosphere.ussa_1976(
        wavelength_range=eval_w[
            (0, -1),
        ]
    )
    template, params = traverse(Scene(objects={"atmosphere": atmosphere}))
    mi_scene: mi.Scene = mi.load_dict(template.render(KernelContext()))

    # Mono evaluation generates valid parameter update tables
    mi_params: mi.SceneParameters = mi.traverse(mi_scene)

    for w in eval_w:
        ctx = KernelContext(si={"w": w})
        mi_params.update(params.render(ctx))


def test_molecular_atmosphere_switches(mode_mono, absorption_dataset_550nm):
    # Absorption can be deactivated
    atmosphere = MolecularAtmosphere(
        absorption_dataset=absorption_dataset_550nm,
        has_absorption=False,
    )
    ctx = KernelContext()
    radprops = atmosphere.eval_radprops(ctx.si, optional_fields=True)
    npt.assert_allclose(radprops.sigma_a, 0.0)

    # Scattering can be deactivated
    skipif_data_not_found(
        f"spectra/absorption/us76_u86_4/us76_u86_4-spectra-18000_19000.nc"
    )
    atmosphere = MolecularAtmosphere(
        absorption_dataset=absorption_dataset_550nm,
        has_scattering=False,
    )
    ctx = KernelContext()
    radprops = atmosphere.eval_radprops(ctx.si, optional_fields=True)
    npt.assert_allclose(radprops.sigma_s, 0.0)

    # At least one must be active
    with pytest.raises(ValueError):
        MolecularAtmosphere(
            absorption_dataset=absorption_dataset_550nm,
            has_absorption=False,
            has_scattering=False,
        )



import numpy as np
import pytest
import xarray as xr

import eradiate.thermoprops.us76 as us76
import eradiate.thermoprops.util as util
from eradiate import unit_registry as ureg


def test_compute_column_number_density():
    # compute correctly (regular altitude mesh)
    ds = xr.Dataset(
        data_vars={
            "mr": (("species", "z_layer"), np.ones((1, 3)), {"units": ""}),
            "n": ("z_layer", np.arange(1, 4), {"units": "m^-3"}),
        },
        coords={
            "z_layer": ("z_layer", np.arange(0, 3) + 0.5, {"units": "m"}),
            "z_level": ("z_level", np.arange(0, 4), {"units": "m"}),
            "species": ("species", ["H2O"], {}),
        },
    )
    assert util.column_number_density(ds, "H2O") == ureg.Quantity(6, "m^-2")

    # compute correctly (irregular altitude mesh)
    ds = xr.Dataset(
        data_vars={
            "mr": (("species", "z_layer"), np.ones((1, 3)), {"units": ""}),
            "n": ("z_layer", np.arange(1, 4), {"units": "m^-3"}),
        },
        coords={
            "z_layer": ("z_layer", [0.5, 2.0, 6.5], {"units": "m"}),
            "z_level": ("z_level", [0.0, 1.0, 3.0, 10.0], {"units": "m"}),
            "species": ("species", ["H2O"], {}),
        },
    )
    assert util.column_number_density(ds, "H2O") == ureg.Quantity(26, "m^-2")

    # compute correctly (mutiple species)
    ds = xr.Dataset(
        data_vars={
            "mr": (("species", "z_layer"), 0.5 * np.ones((2, 3)), {"units": ""}),
            "n": ("z_layer", np.arange(1, 4), {"units": "m^-3"}),
        },
        coords={
            "z_layer": ("z_layer", [0.5, 2.0, 6.5], {"units": "m"}),
            "z_level": ("z_level", [0.0, 1.0, 3.0, 10.0], {"units": "m"}),
            "species": ("species", ["H2O", "O3"], {}),
        },
    )
    assert util.column_number_density(ds, "H2O") == ureg.Quantity(13, "m^-2")
    assert util.column_number_density(ds, "O3") == ureg.Quantity(13, "m^-2")


def test_column_mass_density():
    # returns a quantity.
    ds = xr.Dataset(
        data_vars={
            "mr": (("species", "z_layer"), np.ones((1, 3)), {"units": ""}),
            "n": ("z_layer", np.arange(1, 4), {"units": "m^-3"}),
        },
        coords={
            "z_layer": ("z_layer", np.arange(0, 3) + 0.5, {"units": "m"}),
            "z_level": ("z_level", np.arange(0, 4), {"units": "m"}),
            "species": ("species", ["H2O"], {}),
        },
    )
    assert isinstance(util.column_mass_density(ds, "H2O"), ureg.Quantity)


def test_compute_number_density_at_surface():
    # compute correctly
    ds = xr.Dataset(
        data_vars={
            "mr": (("species", "z_layer"), 0.6 * np.ones((1, 3)), {"units": ""}),
            "n": ("z_layer", np.arange(1, 4), {"units": "m^-3"}),
        },
        coords={
            "z_layer": ("z_layer", np.arange(0, 3) + 0.5, {"units": "m"}),
            "z_level": ("z_level", np.arange(0, 4), {"units": "m"}),
            "species": ("species", ["H2O"], {}),
        },
    )
    value = util.number_density_at_surface(ds, "H2O")
    assert value == ureg.Quantity(0.6, "m^-3")


def test_mass_density_at_surface():
    # returns a quantity
    ds = xr.Dataset(
        data_vars={
            "mr": (("species", "z_layer"), np.ones((1, 3)), {"units": ""}),
            "n": ("z_layer", np.arange(1, 4), {"units": "m^-3"}),
        },
        coords={
            "z_layer": ("z_layer", np.arange(0, 3) + 0.5, {"units": "m"}),
            "z_level": ("z_level", np.arange(0, 4), {"units": "m"}),
            "species": ("species", ["H2O"], {}),
        },
    )
    assert isinstance(util.mass_density_at_surface(ds, "H2O"), ureg.Quantity)


def test_scaling_factor():
    """Returns correct value."""

    assert np.isclose(
        util._scaling_factor(
            initial_amount=1.0 / ureg.m**3,
            target_amount=10.0 / ureg.m**3,
        ),
        10.0,
    )

    assert np.isclose(
        util._scaling_factor(
            initial_amount=1.0 / ureg.m**3,
            target_amount=0.0 / ureg.m**3,
        ),
        0.0,
    )


def test_scaling_factor_invalid():
    """Raises when both initial and target amounts are zero."""
    with pytest.raises(ValueError):
        util._scaling_factor(
            initial_amount=0.0 / ureg.m**3,
            target_amount=1.0 / ureg.m**3,
        )


def test_compute_scaling_factors():
    # compute correctly
    ds = xr.Dataset(
        data_vars={
            "mr": (("species", "z_layer"), 0.5 * np.ones((2, 3)), {"units": ""}),
            "n": ("z_layer", np.arange(1, 4), {"units": "m^-3"}),
        },
        coords={
            "z_layer": ("z_layer", [0.5, 2.0, 6.5], {"units": "m"}),
            "z_level": ("z_level", [0.0, 1.0, 3.0, 10.0], {"units": "m"}),
            "species": ("species", ["H2O", "O3"], {}),
        },
    )

    factors = util.compute_scaling_factors(
        ds=ds, concentration={"H2O": ureg.Quantity(26.0, "m^-2")}
    )
    assert factors["H2O"] == 2.0

    ds = xr.Dataset(
        data_vars={
            "mr": (("species", "z_layer"), 0.5 * np.ones((2, 3)), {"units": ""}),
            "n": ("z_layer", 2.687e20 * np.ones(3), {"units": "m^-3"}),
        },
        coords={
            "z_layer": ("z_layer", [0.5, 2.0, 6.5], {"units": "m"}),
            "z_level": ("z_level", [0.0, 1.0, 3.0, 10.0], {"units": "m"}),
            "species": ("species", ["H2O", "O3"], {}),
        },
    )

    factors = util.compute_scaling_factors(
        ds=ds, concentration={"O3": ureg.Quantity(10, "dobson_units")}
    )
    assert np.isclose(factors["O3"], 2.0, rtol=1e-9)


def test_human_readable():
    assert util.human_readable(["a", "b", "c"]) == "a, b and c"


def test_rescale_concentration():
    ds = xr.Dataset(
        data_vars={
            "mr": (("species", "z_layer"), 0.1 * np.ones((2, 3)), {"units": ""}),
            "n": ("z_layer", 2.687e20 * np.ones(3), {"units": "m^-3"}),
        },
        coords={
            "z_layer": ("z_layer", [0.5, 2.0, 6.5], {"units": "m"}),
            "z_level": ("z_level", [0.0, 1.0, 3.0, 10.0], {"units": "m"}),
            "species": ("species", ["H2O", "O3"], {}),
        },
        attrs=dict(
            history="",
        ),
    )
    initial_ozone_amount = util.column_number_density(ds=ds, species="O3")
    new_ozone_amount = ureg.Quantity(1, "dobson_units")
    factors = util.compute_scaling_factors(
        ds=ds, concentration={"O3": new_ozone_amount}
    )
    rescaled_ds = util.rescale_concentration(ds=ds, factors=factors, inplace=False)
    ds_ozone_amount = util.column_number_density(ds=ds, species="O3")
    ozone_amount = util.column_number_density(ds=rescaled_ds, species="O3")
    assert ds_ozone_amount == initial_ozone_amount
    assert ozone_amount.to("dobson_unit") == new_ozone_amount

    # too large rescaling factors raise error
    new_ozone_amount = ureg.Quantity(10, "dobson_units")
    factors = util.compute_scaling_factors(
        ds=ds, concentration={"O3": new_ozone_amount}
    )
    with pytest.raises(ValueError):
        util.rescale_concentration(ds=ds, factors=factors, inplace=False)


def test_interpolate():
    # column amounts are conserved
    ds = xr.Dataset(
        data_vars={
            "mr": (("species", "z_layer"), 0.1 * np.ones((2, 3)), {"units": ""}),
            "n": ("z_layer", 2.687e20 * np.ones(3), {"units": "m^-3"}),
        },
        coords={
            "z_layer": ("z_layer", [0.5, 2.0, 6.5], {"units": "m"}),
            "z_level": ("z_level", [0.0, 1.0, 3.0, 10.0], {"units": "m"}),
            "species": ("species", ["H2O", "O3"], {}),
        },
        attrs=dict(
            history="",
        ),
    )
    initial_amounts = {s: util.column_number_density(ds, s) for s in ds.species.values}
    interpolated = util.interpolate(
        ds=ds, method="linear", z_level=np.linspace(0, 8, 9), conserve_columns=True
    )
    amounts = {
        s: util.column_number_density(interpolated, s)
        for s in interpolated.species.values
    }
    for s in ds.species.values:
        assert np.isclose(amounts[s], initial_amounts[s], rtol=1e-9)


def test_water_vapor_saturation_pressure():
    # values are correct, reference value from:
    # https://www.engineeringtoolbox.com/water-vapor-saturation-pressure-d_599.html
    t = ureg.Quantity(18, "celsius")
    p = util.water_vapor_saturation_pressure(t=t)
    assert np.isclose(p, ureg.Quantity(2.065, "kPa"), rtol=0.1)

    # accepts temperature below freezing point
    t = ureg.Quantity(-10, "celsius")
    p = util.water_vapor_saturation_pressure(t=t)


def test_equilibrium_water_vapor_fraction():
    # raises when equilibrium does not exist
    with pytest.raises(ValueError):
        util.equilibrium_water_vapor_fraction(
            p=ureg.Quantity(3, "kPa"), t=ureg.Quantity(50, "celsius")
        )
        util.equilibrium_water_vapor_fraction(
            p=ureg.Quantity(3, "kPa"), t=ureg.Quantity(-10, "celsius")
        )
        util.equilibrium_water_vapor_fraction(
            p=ureg.Quantity(100, "Pa"), t=ureg.Quantity(10, "celsius")
        )
        util.equilibrium_water_vapor_fraction(
            p=ureg.Quantity(100, "Pa"), t=ureg.Quantity(-10, "celsius")
        )
        util.equilibrium_water_vapor_fraction(
            p=ureg.Quantity(1, "bar"), t=ureg.Quantity(120, "celsius")
        )
    # values are in [0, 1]
    value = util.equilibrium_water_vapor_fraction(
        p=ureg.Quantity(90, "kPa"), t=ureg.Quantity(20, "celsius")
    )
    assert 0.0 <= value <= 1.0


def test_find_regular_params_gcd():
    with pytest.raises(ValueError):
        util._find_regular_params_gcd([1, 2, 3], -1.0)

    with pytest.raises(ValueError):
        util._find_regular_params_gcd([1, 2, 3], 0.0)

    assert util._find_regular_params_gcd(np.arange(10)) == (10, 1)

    mesh = [0.0, 1200.0, 2500.0, 5300.0, 10000.0, 25500.0]
    assert util._find_regular_params_gcd(mesh) == (256, 100)
    assert util._find_regular_params_gcd(mesh, unit_number=1000.0) == (26, 1000)


def test_to_regular():
    # check that the bounds are kept the same
    mesh = [0.0, 1200.0, 2500.0, 5300.0, 10000.0, 25500.0]
    regular = util._to_regular(mesh, atol=1000)
    assert regular[0] == mesh[0]
    assert regular[-1] == mesh[-1]


# TODO: test further
def test_make_profile_regular():
    irregular_profile = us76.make_profile(
        levels=ureg.Quantity([0.0, 2.0, 8.0, 10.0, 16.0, 23.0, 32.0, 47.0, 50.0], "km")
    )

    regular_profile = util.make_profile_regular(irregular_profile, atol=1000.0)



from __future__ import annotations

import pint

# This table hard-codes a wavenumber bin index for spectral data.
# Beware: this is fragile, it can break if not carefully maintained!
WAVENUMBER_BINS = {
    "spectra/absorption/us76_u86_4/us76_u86_4-spectra": [
        (4000, 5000),
        (5000, 6000),
        (6000, 7000),
        (7000, 8000),
        (8000, 9000),
        (9000, 10000),
        (10000, 11000),
        (11000, 12000),
        (12000, 13000),
        (13000, 14000),
        (14000, 15000),
        (15000, 16000),
        (16000, 17000),
        (17000, 18000),
        (18000, 19000),
        (19000, 20000),
        (20000, 21000),
        (21000, 22000),
        (22000, 23000),
        (23000, 24000),
        (24000, 25000),
        (25000, 25711),
    ]
}


def resolve_bin(group: str, wavenumber: pint.Quantity) -> tuple[int, int]:
    """
    Return the wavenumber bin corresponding to a group and a wavenumber value,
    if it exists.
    """
    wavenumber_m = wavenumber.m_as("cm^-1")

    for w_bin in WAVENUMBER_BINS[group]:
        w_min, w_max = w_bin
        if w_min <= wavenumber_m <= w_max:
            return w_bin

    raise ValueError(
        "Cannot find wavenumber bin corresponding to wavenumber "
        f"value {wavenumber} for group {group}"
    )


def get_us76_u86_4_spectrum_filename(wavelength: pint.Quantity):
    prefix = "spectra/absorption/us76_u86_4/us76_u86_4-spectra"
    wn_min, wn_max = resolve_bin(prefix, wavenumber=1.0 / wavelength)
    return f"{prefix}-{wn_min}_{wn_max}.nc"



import numpy as np
import pytest
import xarray as xr

from eradiate.thermoprops.afgl_1986 import make_profile
from eradiate.thermoprops.util import (
    column_number_density,
    number_density_at_surface,
)
from eradiate.units import unit_registry as ureg


@pytest.mark.parametrize(
    "model_id",
    [
        "midlatitude_summer",
        "midlatitude_winter",
        "subarctic_summer",
        "subarctic_summer",
        "subarctic_winter",
        "us_standard",
    ],
)
def test_make_profile(mode_mono, model_id):
    """
    All thermophysical properties profile can be made.
    """
    ds = make_profile(model_id=model_id)
    assert isinstance(ds, xr.Dataset)


@pytest.mark.parametrize(
    "model_id",
    [
        "midlatitude_summer",
        "midlatitude_winter",
        "subarctic_summer",
        "subarctic_summer",
        "subarctic_winter",
        "us_standard",
    ],
)
@pytest.mark.parametrize(
    "levels",
    [
        ureg.Quantity(np.linspace(0, 120, 121), "km"),
        ureg.Quantity(np.linspace(0, 86, 87), "km"),
        ureg.Quantity(np.linspace(0, 40, 41), "km"),
    ],
)
def test_make_profile_levels(mode_mono, model_id, levels):
    """
    Atmospheric thermophysical profile altitude levels array has the same
    shape as the input 'levels'.
    """
    ds = make_profile(model_id=model_id, levels=levels)
    assert ds.z_level.shape == levels.shape


@pytest.mark.parametrize(
    "model_id",
    [
        "midlatitude_summer",
        "midlatitude_winter",
        "subarctic_summer",
        "subarctic_summer",
        "subarctic_winter",
        "us_standard",
    ],
)
@pytest.mark.parametrize(
    "levels",
    [
        ureg.Quantity(np.linspace(0, 120, 121), "km"),
        ureg.Quantity(np.linspace(0, 86, 87), "km"),
        ureg.Quantity(np.linspace(0, 40, 41), "km"),
    ],
)
def test_make_profile_concentrations(mode_mono, model_id, levels):
    """
    Concentration-rescaled profile concentrations match the input concentrations.
    """
    concentrations = {
        "H2O": ureg.Quantity(5e23, "m^-2"),
        "O3": ureg.Quantity(0.5, "dobson_unit"),
        "CO2": ureg.Quantity(400e-6, ""),
    }
    ds = make_profile(model_id=model_id, levels=levels, concentrations=concentrations)

    column_amount_H2O = column_number_density(ds=ds, species="H2O")
    column_amount_O3 = column_number_density(ds=ds, species="O3")
    surface_amount_CO2 = ds.mr.sel(species="CO2").values[0]

    assert np.isclose(column_amount_H2O, concentrations["H2O"], rtol=1e-9)
    assert np.isclose(column_amount_O3, concentrations["O3"], rtol=1e-9)
    assert np.isclose(surface_amount_CO2, concentrations["CO2"], rtol=1e-9)


@pytest.mark.parametrize(
    "model_id",
    [
        "midlatitude_summer",
        "midlatitude_winter",
        "subarctic_summer",
        "subarctic_summer",
        "subarctic_winter",
        "us_standard",
    ],
)
@pytest.mark.parametrize(
    "levels",
    [
        ureg.Quantity(np.linspace(0, 120, 121), "km"),
        ureg.Quantity(np.linspace(0, 86, 87), "km"),
        ureg.Quantity(np.linspace(0, 40, 41), "km"),
    ],
)
def test_make_profile_concentrations_invalid(mode_mono, model_id, levels):
    """
    Too large concentrations raise.
    """
    with pytest.raises(ValueError):
        make_profile(
            model_id=model_id,
            levels=levels,
            concentrations={"CO2": ureg.Quantity(400, "")},
        )
